#!/bin/sh

#
# This script makes easier using the NS3D code.
# It performs all the necessary steps:
#     - copying the necessary files (input data, sources...) into the execution directory
#     - compilation
#     - execution
#     - save of the results in the save directory
#

#____________________________________________________________
#
#    PATHS/NAMES
#____________________________________________________________
#

# ------------------------------
# Directories/files used
# ------------------------------

# directory where the results are saved (a sub-directory is automatically created for each run)
dir_save="/toon/usr/deloncle/test"

# directory where the source files are
dir_source="../source"

# working directory where the code runs (a sub-directory is automatically created for each run)
dir_work="/toon/usr/deloncle/NS3D_work"

# optional paths to the input data files (previous results data file to restart a run, initial velocity, base flow)
velocity_init=""
base2d_init=""

# name of the simulation, useful to determine the name of the working/save folders
run_name="test_speed"

# Makefile to use for the compilation (examples are provided in "source/Makefile_examples")
# if this field is blank, an automatic selection is implemented in the script code (see directly below...)
makefile_file="../source/Makefile_examples/Makefile_Intel64_nofft"

# Number of parallel OpenMP processes (0 means non-OpenMP code)
nb_openmp_processes=0


#____________________________________________________________
#
#    COMPILATION PARAMETERS (config.h)
#____________________________________________________________
#

mkdir -p temp
cat > temp/config.h_pid$$ << EOF
! *******************************************
!    NS3D: compilation parameters
! *******************************************

! This file is used by the pre-processor during the compilation step.

! Avoid multiple inclusions of this file. Do not change!
#ifndef CONFIG_H
#define CONFIG_H

! Number of colocation points
#define DIMX 256
#define DIMY 256
#define DIMZ 128

! Padding along each direction
#define PPADKX 0
#define PPADKY 0
#define PPADKZ 0

! The following time schemes are available:
!     - AB2: Adams-Bashforth of order 2
!     - RK2: Runge and Kutta of order 2
!     - RK3: Runge and Kutta of order 3
!     - RK4: Runge and Kutta of order 4
#define RK4

! We can simulate a single complex kz mode for a perturbative and linearized run.
! Warning: for correct results, the eigenmode growhtrate must be purely real.
#undef SINGLE_MODE

! we estimate a translation velocity of a basic state
#undef calcul_ut

#endif
EOF



#____________________________________________________________
#
#    SIMULATION PARAMETERS (data.in)
#____________________________________________________________
#

cat > temp/data.in_pid$$ << EOF
********************************************
*    NS3D: simulation parameters           *
********************************************

This file is read at every simulation start.

*** discretization variables ***
lx___________________________   1.
ly___________________________   1.
lz___________________________   1.
dt___________________________   0.01
begin________________________   0.
itmax________________________   5
de_aliasing__________________   T
    squared:1_or_elliptic:2__   1
    radius_truncation_x______   0.99
    radius_truncation_y______   0.99
    radius_truncation_z______   1.E30

*** physical variables ***
viscosity____________________   2.E-5
stratified___________________   T
    brunt_vaisala_frequency__   10.
    schmidt_number___________   1.
2omega_______________________   0.

*** Type of simulation ***
perturbative_________________   F
    linear___________________   T

*** Base state (only perturbation run) ***
base2D_type__________________   null

*** Initial velocity ***
velo_type____________________   tanh
white_noise__________________   0.

*** Output ***
output1_period_______________   1
output2_period_______________   0
output3_period_______________   0

EOF


#____________________________________________________________
#
#     DO NOT MODIFY BELOW!
#____________________________________________________________
#

printf "\n********** START OF THE NS3D SCRIPT *********\n"


#_____________________________________________________
#
#    PRELIMINARY TASKS
#_____________________________________________________
#

# we read some simulation parameters in the configuration files
NX=`grep "#define DIMX" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMX/ {print $3}'`
NY=`grep "#define DIMY" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMY/ {print $3}'`
NZ=`grep "#define DIMZ" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMZ/ {print $3}'`
LX=`awk '/^lx/ {print $2}' temp/data.in_pid$$`
LY=`awk '/^ly/ {print $2}' temp/data.in_pid$$`
LZ=`awk '/^lz/ {print $2}' temp/data.in_pid$$`
Dt=`awk '/^dt/ {print $2}' temp/data.in_pid$$`
nu=`awk '/^viscosity/ {print $2}' temp/data.in_pid$$`
stratifie=`awk '/^stratified/ {print $2}' temp/data.in_pid$$`
brunt=`awk '/brunt_vaisala_frequency/ {print $2}' temp/data.in_pid$$`

if [ "$stratifie" = "F" ]; then
    brunt=0
fi

# if the makefile_file variable is not defined, we search for a suitable Makefile
if [ -z "$makefile_file" ]; then
    printf "No Makefile specified...\n"

    if [ -n "`echo $HOSTNAME | grep -e raspoutine -e dracula -e joebar`" ]; then
        makefile_file=${dir_source}/Makefile_examples/Makefile_Intel64_fftw
    elif [ -n "`echo $HOSTNAME | grep -e tetsuo -e kelvin -e mandrake -e jill -e gailuron`" ]; then
        makefile_file=${dir_source}/Makefile_examples/Makefile_Intel32_fftw
    elif [ -n "`echo $HOSTNAME | grep -e rahan`" ]; then
        makefile_file=${dir_source}/Makefile_examples/Makefile_Mac_fftw
    elif [ -n "`echo $HOSTNAME | grep -e gibi -e melusine`" ]; then
        makefile_file=${dir_source}/Makefile_examples/Makefile_SGI_fftw
    fi
fi

# we use the OpenMP version of the Makefile if necessary
if [ $nb_openmp_processes -gt 0 ] ; then
    makefile_file=${makefile_file}_OpenMP
fi

printf "Makefile: ${makefile_file}\n"

sed "s/-DFFTW_OPENMP=[0-9]*/-DFFTW_OPENMP=$nb_openmp_processes/" $makefile_file > temp/Makefile_pid$$


#_____________________________________________________
#
#    COPY OF THE FILES IN THE WORKING DIRECTORY
#_____________________________________________________
#

printf "Copy of the files in the working directory...\n"

# we define the full name of the working directory and create it
date_heure=`date +%d-%m-%y_%Hh%M`
dir_work_full=${dir_work}/ns3d_work_${run_name}_${NX}x${NY}x${NZ}_L=${LX}x${LY}x${LZ}_Dt=${Dt}_nu=${nu}_N=${brunt}_${date_heure}_pid$$

mkdir -p $dir_work_full

# we copy all the necessary files (sources, data file, configuration file) in the working directory
cp -r  $dir_source/*                         $dir_work_full/
cp -f  temp/Makefile_pid$$                   $dir_work_full/Makefile
mv -f  temp/config.h_pid$$                   $dir_work_full/config.h
mv -f  temp/data.in_pid$$                    $dir_work_full/data.in

# we copy the input data files if necessary
if [ -f "$base2d_init" ]; then
    cp    $base2d_init        $dir_work_full/base2D.init
fi

if [ -f "$velocity_init" ]; then
    cp    $velocity_init      $dir_work_full/velocity.init
fi


#_____________________________________________________
#
#    COMPILATION
#_____________________________________________________
#

printf "Compilation...\n"

# we compile
cd $dir_work_full
make ns3d
chmod 744 *

#_____________________________________________________
#
#    LAUNCHING THE RUN
#_____________________________________________________
#

# we print everything in the file "listing.txt"
printf "\nStart of the simulation on $HOSTNAME: `date '+%A %d %B %Y a %T'`\n" | tee -a listing.txt
echo "*************************************************************************" >> listing.txt

echo " " >> listing.txt
echo " " >> listing.txt

echo "COMPILATION PARAMETERS" >> listing.txt
echo "***************************" >> listing.txt
cat config.h >> listing.txt
echo " " >> listing.txt
cat extended_config.h >> listing.txt

echo " " >> listing.txt
echo " " >> listing.txt

echo "SIMULATION PARAMETERS" >> listing.txt
echo "***************************" >> listing.txt
cat data.in >> listing.txt

echo " " >> listing.txt
echo " " >> listing.txt

echo " "
ulimit -s unlimited
if [ "$1" = "-s" ] ; then
    time nice -19 ./ns3d | tee -a listing.txt
else
    time nice -19 ./ns3d >> listing.txt
fi

echo " " >> listing.txt
echo " " >> listing.txt

printf "\nEnd of the simulation: `date '+%A %d %B %Y a %T'`\n" | tee -a listing.txt
echo "*************************************************************************" >> listing.txt

#_____________________________________________________
#
#     SAVING THE RESULTS
#_____________________________________________________
#

printf "\nSave of the results...\n"

# we create the saving directory
dir_save_full=$dir_save/ns3d_${run_name}_${NX}x${NY}x${NZ}_L=${LX}x${LY}x${LZ}_Dt=${Dt}_nu=${nu}_N=${brunt}_$date_heure

if [ -d "$dir_save_full" ] ; then
    dir_save_full=$dir_save_full"_"pid$$
fi

mkdir -p $dir_save_full

# we delete the executable, the object files and the module files. We save all the other files.
rm -f ns3d *.mod *.o base2D.init
mv * $dir_save_full

# we delete the working directory
cd ..
rm -rf $dir_work_full

is_empty=`ls -A $dir_work`
if [ -z "$is_empty" ]; then
    cd ..
    rm -rf $dir_work;
fi

printf "\n********** END OF THE NS3D SCRIPT *********\n"

exit
