#!/bin/sh

# I modified this script compared to job_brodie.sh to avoid uploading a big initial velocity field and to copy it directly from gaya

#
# This script makes easier using the NS3D code on the NEC-SX8 (brodie) of IDRIS
# It performs all the necessary steps:
#     - copying the necessary files (input data, sources...) into the execution directory
#     - compilation
#     - execution
#     - save of the results in the save directory
#
# It must be run from the computer connected to the exterior "griffu"

#____________________________________________________________
#
#    PATHS/NAMES
#____________________________________________________________
#

# ------------------------------
# Directories/files used
# ------------------------------

# directory where the results are saved on Gaya (a sub-directory is automatically created for each run)
dir_save_gaya="/u/rech/bll/rbll200/simple"

# directory where the source files are
dir_source="../source"

# optional paths to the input data files (initial velocity, base flow).
# "velocity_init" and "base2d_init" corresponds to files that are initially on griffu.
# "velocity_init_gaya" corresponds to a file initially on Gaya
velocity_init=""
base2d_init="/toon/usr/deloncle/linear_stability_vortex_arrays/simple/base_state/ns3d_simple_a_1_256x1024x1_L=30.x120.x1._Dt=0.02_nu=2.E-5_N=0_03-10-07_15h37/base2D_rotation_solide_512x2048_L=30x120_be"
velocity_init_gaya="/u/rech/bll/rbll200/simple/ns3d_simple_512x2048x1_L=30.x120.x0._Dt=0.01_nu=2.E-5_N=0_brodie_03-10-07_18h12_pid10054/velo_rho_vort.t=1200.000"

# name of the simulation, useful to determine the name of the working/save folders
run_name="simple"

# Number of parallel (OpenMP and MPI) processes (0 means non-parallel code)
nb_openmp_processes=0
nb_mpi_processes=0

# Number of recursive submission of the script. It is possible to perform several runs in a row
# using the final velocity of the previous run as initial velocity (0 or 1 means single run)
nb_recursive_run=0

# In batch mode, it is necessary to indicate some additional parameters
#   - name of the job (optional)
#   - cpu time=10:00:00  # max time/process (HH:MM:SS)
#   - job time=10:00:00  # max time/job (HH:MM:SS). Different from cpu time for MPI code.
#   - memory used and the unit used (Mb, Gb) for the job.
name_nqs="ns3d"
time_cpu_nqs="10:00:00"
time_job_nqs="10:00:00"
memory_job_nqs="6Gb"


#____________________________________________________________
#
#    COMPILATION PARAMETERS (config.h)
#____________________________________________________________
#

mkdir -p temp
cat > temp/config.h_pid$$ << EOF
! *******************************************
!    NS3D: compilation parameters
! *******************************************

! This file is used by the pre-processor during the compilation step.

! Avoid multiple inclusions of this file. Do not change!
#ifndef CONFIG_H
#define CONFIG_H

! Number of colocation points
#define DIMX 512
#define DIMY 2048
#define DIMZ 1

! Padding along each direction
#define PPADKX 0
#define PPADKY 0
#define PPADKZ 0

! The following time schemes are available:
!     - AB2: Adams-Bashforth of order 2
!     - RK2: Runge and Kutta of order 2
!     - RK3: Runge and Kutta of order 3
!     - RK4: Runge and Kutta of order 4
#define RK4

! We can simulate a single complex kz mode for a perturbative and linearized run.
! Warning: for correct results, the eigenmode growhtrate must be purely real.
#define SINGLE_MODE

! we estimate a translation velocity of a basic state
#undef calcul_ut

#endif
EOF



#____________________________________________________________
#
#    SIMULATION PARAMETERS (data.in)
#____________________________________________________________
#

cat > temp/data.in_pid$$ << EOF
********************************************
*    NS3D: simulation parameters           *
********************************************

This file is read at every simulation start.

** discretization variables ***
lx___________________________   30.
ly___________________________   120.
lz___________________________   0.
dt___________________________   0.01
begin________________________   0.
itmax________________________   120000
de_aliasing__________________   T
    squared:1_or_elliptic:2__   1
    radius_truncation_x______   0.99
    radius_truncation_y______   0.99
    radius_truncation_z______   1.E30

*** physical variables ***
viscosity____________________   2.E-5
stratified___________________   F
    brunt_vaisala_frequency__   10.
    schmidt_number___________   1.
2omega_______________________   0.

*** Type of simulation ***
perturbative_________________   T
    linear___________________   T

*** Base state (only perturbation run) ***
base2D_type__________________   file

*** Initial velocity ***
velo_type____________________   file
white_noise__________________   0.

*** Output ***
output1_period_______________   1000
output2_period_______________   0
output3_period_______________   0


*** additionnal variables... ***

* Stuart vortices *
concentration_stuart_________   0.25

* File gaussian vortices *
file_nb_vortices_____________   2
Vortex 1
    position_x_______________   7.5
    position_y_______________   15.
    circulation______________   6.28
    core_radius______________   1.
Vortex 2
    position_x_______________   22.5
    position_y_______________   15.
    circulation______________   6.28
    core_radius______________   1.

* Random gaussian vortices *
rnd_nb_vortices______________   10
rnd_min_distance_____________   1.
rnd_mean_gamma_______________   6.28
rnd_std_gamma________________   0.
rnd_mean_radius______________   1.
rnd_std_radius_______________   0.

EOF


#____________________________________________________________
#
#     DO NOT MODIFY BELOW!
#____________________________________________________________
#

printf "\n********** START OF THE NS3D SCRIPT *********\n"


#_____________________________________________________
#
#    PRELIMINARY TASKS
#_____________________________________________________
#

# useful to personalize the names of the used directories
date_hour=`date +"%d-%m-%y_%Hh%M"`
pid_job=$$

# we read some simulation parameters in the configuration files
NX=`grep "#define DIMX" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMX/ {print $3}'`
NY=`grep "#define DIMY" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMY/ {print $3}'`
NZ=`grep "#define DIMZ" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMZ/ {print $3}'`
LX=`awk '/^lx/ {print $2}' temp/data.in_pid$$`
LY=`awk '/^ly/ {print $2}' temp/data.in_pid$$`
LZ=`awk '/^lz/ {print $2}' temp/data.in_pid$$`
Dt=`awk '/^dt/ {print $2}' temp/data.in_pid$$`
nu=`awk '/^viscosity/ {print $2}' temp/data.in_pid$$`
stratifie=`awk '/^stratified/ {print $2}' temp/data.in_pid$$`
brunt=`awk '/brunt_vaisala_frequency/ {print $2}' temp/data.in_pid$$`

if [ "$stratifie" = "F" ]; then
    brunt=0
fi

dir_name="${run_name}_${NX}x${NY}x${NZ}_L=${LX}x${LY}x${LZ}_Dt=${Dt}_nu=${nu}_N=${brunt}"


# we determine the Makefile to use depending on the number of MPI and OpenMP processes
if [ $nb_openmp_processes -gt 0 ] ; then
    makefile_file=${dir_source}/Makefile_examples/Makefile_NEC_MATHKEISAN_OpenMP_LARGE_FILE
elif [ $nb_mpi_processes -gt 0 ] ; then
    makefile_file=${dir_source}/Makefile_examples/Makefile_NEC_MATHKEISAN_MPI
else
    makefile_file=${dir_source}/Makefile_examples/Makefile_NEC_MATHKEISAN
fi

printf "Makefile: ${makefile_file}\n"

#_____________________________________________________
#
#    MODIFICATION OF THE EXECUTION SCRIPT "job_brodie_exec"
#_____________________________________________________
#

if [ $nb_recursive_run -gt 1 ] ; then
    nb_recursive_run_left=`expr $nb_recursive_run - 1`
else
    nb_recursive_run_left=0
fi

sed -e s\%^dir_save_gaya=.*%dir_save_gaya=$dir_save_gaya% \
    -e "s/^nb_mpi_processes=.*/nb_mpi_processes=$nb_mpi_processes/" \
    -e "s/^nb_openmp_processes=.*/nb_openmp_processes=$nb_openmp_processes/" \
    -e "s/^date_hour=.*/date_hour=$date_hour/" \
    -e "s/^pid_job=.*/pid_job=$pid_job/" \
    -e "s/^dir_name=.*/dir_name=$dir_name/" \
    -e "s/^nb_recursive_run_left=.*/nb_recursive_run_left=$nb_recursive_run_left/" \
    job_brodie_exec.in > temp/job_brodie_exec_pid$$

# we perform the necessary modifications for a parallel code
sed -e "s/-DMPI=[0-9]*/-DMPI=$nb_mpi_processes/" \
    $makefile_file > temp/Makefile_pid$$

# we perform the necessary modifications for a parallel code
if [ $nb_openmp_processes -gt 0 ] ; then
    nb_processes=$nb_openmp_processes
elif [ $nb_mpi_processes -gt 0 ] ; then
    nb_processes=$nb_mpi_processes
else
    sed -e "/^#PBS -q multi.*/d" \
        -e "/^#PBS -l cpunum_job=.*/d" \
    temp/job_brodie_exec_pid$$ > temp/job_brodie_exec_pid$$_bis ;

    mv temp/job_brodie_exec_pid$$_bis temp/job_brodie_exec_pid$$ ;
fi

sed -e "/^#!\/bin\/ksh.*/d" \
    -e "\%^. /etc/profile > /dev/null 2>&1.*%d" \
    -e "s/time_cpu_nqs/$time_cpu_nqs/g" \
    -e "s/time_job_nqs/$time_job_nqs/g" \
    -e "s/memory_job_nqs/$memory_job_nqs/g" \
    -e "s/name_nqs/$name_nqs/g" \
    -e "s/nb_processes_ll/$nb_processes/g" \
    temp/job_brodie_exec_pid$$ > temp/job_brodie_exec_pid$$_bis ;

mv temp/job_brodie_exec_pid$$_bis temp/job_brodie_exec_pid$$ ;


#_____________________________________________________
#
#       UPLOAD OF THE FILES ON BRODIE COMPILATION ON BRODIE
#_____________________________________________________
#

printf "\nUpload of the files on Brodie...\n"

# we create a directory on gaya to copy the executable after the compilation on Brodie.
dir_save_gaya_upload=$dir_save_gaya/ns3d_upload_${date_hour}_pid$$
ssh -x rbll200@gaya.idris.fr "mkdir -p $dir_save_gaya_upload"

# we compile in the $WORKDIR of Brodie
printf "\nTransfer of the files on Brodie for the compilation...\n"
printf "Directory of compilation: /workdir/rech/bll/rbll200/ns3d_compil_brodie_pid$$\n"

ssh -x rbll200@brodie.idris.fr "mkdir -p /workdir/rech/bll/rbll200/ns3d_compil_brodie_pid$$"

scp -r     $dir_source/*                       rbll200@brodie.idris.fr:/workdir/rech/bll/rbll200/ns3d_compil_brodie_pid$$
scp        temp/config.h_pid$$                 rbll200@brodie.idris.fr:/workdir/rech/bll/rbll200/ns3d_compil_brodie_pid$$/config.h
scp        temp/data.in_pid$$                  rbll200@brodie.idris.fr:/workdir/rech/bll/rbll200/ns3d_compil_brodie_pid$$/data.in
scp        temp/Makefile_pid$$                 rbll200@brodie.idris.fr:/workdir/rech/bll/rbll200/ns3d_compil_brodie_pid$$/Makefile


printf "\nCompilation on Brodie...\n"
# we load the right version of the MATHKEISAN library (temporary until IDRIS installs the new version)
ssh -x rbll200@brodie.idris.fr ". /etc/profile > /dev/null 2>&1 ; cd /workdir/rech/bll/rbll200/ns3d_compil_brodie_pid$$ ; make ; mfput ns3d  $dir_save_gaya_upload ; cd .. ; rm -rf ns3d_compil_brodie_pid$$"




#_____________________________________
#
#       UPLOAD OF THE FILES ON GAYA
#_____________________________________
#

printf "\nUpload of the files on Gaya...\n"

# we upload all the necessary files (sources, data file, configuration file) in the working directory
rcp -r     $dir_source/*                       rbll200@gaya.idris.fr:$dir_save_gaya_upload
rcp        temp/config.h_pid$$                 rbll200@gaya.idris.fr:$dir_save_gaya_upload/config.h
rcp        temp/data.in_pid$$                  rbll200@gaya.idris.fr:$dir_save_gaya_upload/data.in
rcp        temp/Makefile_pid$$                 rbll200@gaya.idris.fr:$dir_save_gaya_upload/Makefile


# we upload the input data files if necessary
if [ -f "$velocity_init" ]; then
    printf "\n   Copy of the initial velocity from griffu to Gaya...\n"
    rcp -v   $velocity_init         rbll200@gaya.idris.fr:$dir_save_gaya_upload/velocity.init
fi

if [ -f "$base2d_init" ]; then
    printf "\n   Copy of the base state from griffu to Gaya...\n"
    rcp -v   $base2d_init           rbll200@gaya.idris.fr:$dir_save_gaya_upload/base2D.init
fi

if [ $velocity_init_gaya ]; then
    printf "\n   Copy of the initial velocity from Gaya to Gaya...\n"
    ssh -x rbll200@gaya.idris.fr "cp -p $velocity_init_gaya $dir_save_gaya_upload/velocity.init"
fi



#_______________________________________________
#
#       UPLOAD OF THE EXECUTION SCRIPT ON BRODIE
#_______________________________________________
#

printf "\nUpload of the execution script on Brodie...\n"

# we create a directory to save the execution script and we upload it
ssh -x rbll200@brodie.idris.fr "mkdir -p ns3d_job_brodie"
scp temp/job_brodie_exec_pid$$ "rbll200@brodie.idris.fr:ns3d_job_brodie/job_brodie_exec_pid$$"

# we delete the useless configuration files
rm temp/config.h_pid$$ temp/data.in_pid$$ temp/job_brodie_exec_pid$$



#_____________________________________________________
#
#       EXECUTION / SUBMISSION OF THE JOB
#_____________________________________________________
#

# we are in batch mode: we submit the execution script
printf "\nSubmission of the job with NQS...\n"
ssh -x rbll200@brodie.idris.fr ". /etc/profile > /dev/null 2>&1 ; cd ns3d_job_brodie ; chmod 744 job_brodie_exec_pid$$ ; qsub job_brodie_exec_pid$$"

printf "\n********** END OF THE NS3D SCRIPT *********\n"

exit
