#!/bin/sh

#
# This script makes easier using the NS3D code on the Altix 450 SMP (tournesol) of LadHyX
# It performs all the necessary steps:
#     - copying the necessary files (input data, sources...) into the execution directory
#     - compilation
#     - execution
#     - save of the results in the save directory
#
# It must be run from "griffu"

#____________________________________________________________
#
#    PATHS/NAMES
#____________________________________________________________
#

# ------------------------------
# Directories/files used
# ------------------------------

# directory where the results are saved on tournesol (a sub-directory is automatically created for each run)
dir_save_tournesol="/home/deloncle/von_karman_kappa_0.2/Fh_1.2"

# directory where the source files are
dir_source="../source"

# optional paths to the input data files (initial velocity, base flow).
# "velocity_init" and "base2d_init" corresponds to files that are initially on griffu.
# "velocity_init_tournesol" corresponds to a file initially on tournesol
velocity_init=""
base2d_init="/toon/usr/deloncle/base2D.init"
velocity_init_tournesol=""
base2d_init_tournesol=""

# name of the simulation, useful to determine the name of the working/save folders
run_name="karman_kappa_0.2"

# Number of parallel (OpenMP and MPI) processes (0 means non-parallel code)
nb_openmp_processes=0
nb_mpi_processes=0

# Number of recursive submission of the script. It is possible to perform several runs in a row
# using the final velocity of the previous run as initial velocity (0 or 1 means single run)
nb_recursive_run=0

# In batch mode, it is necessary to indicate some additional parameters
#   - name of the job (optional)
#   - job time=10:00:00  # max time/job (HH:MM:SS)
#   - memory used and the unit used (Mb, Gb) for the job.
name_nqs="ns3d"
time_job_nqs="99:00:00"
memory_job_nqs="1Gb"


#____________________________________________________________
#
#    COMPILATION PARAMETERS (config.h)
#____________________________________________________________
#

mkdir -p temp
cat > temp/config.h_pid$$ << EOF
! *******************************************
!    NS3D: compilation parameters
! *******************************************

! This file is used by the pre-processor during the compilation step.

! Avoid multiple inclusions of this file. Do not change!
#ifndef CONFIG_H
#define CONFIG_H

! Number of colocation points
#define DIMX 256
#define DIMY 640
#define DIMZ 1

! Padding along each direction
#define PPADKX 0
#define PPADKY 0
#define PPADKZ 0

! The following time schemes are available:
!     - AB2: Adams-Bashforth of order 2
!     - RK2: Runge and Kutta of order 2
!     - RK3: Runge and Kutta of order 3
!     - RK4: Runge and Kutta of order 4
#define RK4

! We can simulate a single complex kz mode for a perturbative and linearized run.
! Warning: for correct results, the eigenmode growhtrate must be purely real.
#define SINGLE_MODE

! we estimate a translation velocity of a basic state
#undef calcul_ut

#endif
EOF



#____________________________________________________________
#
#    SIMULATION PARAMETERS (data.in)
#____________________________________________________________
#

cat > temp/data.in_pid$$ << EOF
********************************************
*    NS3D: simulation parameters           *
********************************************

This file is read at every simulation start.

*** discretization variables ***
lx___________________________   15.
ly___________________________   38.
lz___________________________   13.30
dt___________________________   0.01
begin________________________   0.
itmax________________________   120000
de_aliasing__________________   T
    squared:1_or_elliptic:2__   1
    radius_truncation_x______   0.99
    radius_truncation_y______   0.99
    radius_truncation_z______   1.E30

*** physical variables ***
viscosity____________________   2.E-5
stratified___________________   T
    brunt_vaisala_frequency__   0.8333
    schmidt_number___________   1.
2omega_______________________   0.

*** Type of simulation ***
perturbative_________________   T
    linear___________________   T

*** Base state (only perturbation run) ***
base2D_type__________________   file

*** Initial velocity ***
velo_type____________________   null
white_noise__________________   0.000001

*** Output ***
output1_period_______________   1000
output2_period_______________   0
output3_period_______________   0


EOF


#____________________________________________________________
#
#     DO NOT MODIFY BELOW!
#____________________________________________________________
#

printf "\n********** START OF THE NS3D SCRIPT *********\n"


#_____________________________________________________
#
#    PRELIMINARY TASKS
#_____________________________________________________
#

# useful to personalize the names of the used directories
date_hour=`date +"%d-%m-%y_%Hh%M"`
pid_job=$$

# we read some simulation parameters in the configuration files
NX=`grep "#define DIMX" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMX/ {print $3}'`
NY=`grep "#define DIMY" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMY/ {print $3}'`
NZ=`grep "#define DIMZ" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMZ/ {print $3}'`
LX=`awk '/^lx/ {print $2}' temp/data.in_pid$$`
LY=`awk '/^ly/ {print $2}' temp/data.in_pid$$`
LZ=`awk '/^lz/ {print $2}' temp/data.in_pid$$`
Dt=`awk '/^dt/ {print $2}' temp/data.in_pid$$`
nu=`awk '/^viscosity/ {print $2}' temp/data.in_pid$$`
stratifie=`awk '/^stratified/ {print $2}' temp/data.in_pid$$`
brunt=`awk '/brunt_vaisala_frequency/ {print $2}' temp/data.in_pid$$`

if [ "$stratifie" = "F" ]; then
    brunt=0
fi

dir_name="${run_name}_${NX}x${NY}x${NZ}_L=${LX}x${LY}x${LZ}_Dt=${Dt}_nu=${nu}_N=${brunt}"


# we determine the Makefile to use depending on the number of MPI and OpenMP processes
if [ $nb_openmp_processes -gt 0 ] ; then
    makefile_file=${dir_source}/Makefile_examples/Makefile_tournesol_fftw_OpenMP
elif [ $nb_mpi_processes -gt 0 ] ; then
    makefile_file=${dir_source}/Makefile_examples/Makefile_tournesol_fftw_MPI
else
    makefile_file=${dir_source}/Makefile_examples/Makefile_tournesol_fftw
fi

printf "Makefile: ${makefile_file}\n"

#_____________________________________________________
#
#    MODIFICATION OF THE EXECUTION SCRIPT "job_tournesol_exec"
#_____________________________________________________
#

if [ $nb_recursive_run -gt 1 ] ; then
    nb_recursive_run_left=`expr $nb_recursive_run - 1`
else
    nb_recursive_run_left=0
fi

sed -e s\%^dir_save_tournesol=.*%dir_save_tournesol=$dir_save_tournesol% \
    -e "s/^nb_mpi_processes=.*/nb_mpi_processes=$nb_mpi_processes/" \
    -e "s/^nb_openmp_processes=.*/nb_openmp_processes=$nb_openmp_processes/" \
    -e "s/^date_hour=.*/date_hour=$date_hour/" \
    -e "s/^pid_job=.*/pid_job=$pid_job/" \
    -e "s/^dir_name=.*/dir_name=$dir_name/" \
    -e "s/^nb_recursive_run_left=.*/nb_recursive_run_left=$nb_recursive_run_left/" \
    job_tournesol_exec.in > temp/job_tournesol_exec_pid$$

# we perform the necessary modifications for a parallel code
sed -e "s/-DMPI=[0-9]*/-DMPI=$nb_mpi_processes/" \
    -e "s/-DFFTW_OPENMP=[0-9]*/-DFFTW_OPENMP=$nb_openmp_processes/" \
    $makefile_file > temp/Makefile_pid$$

# we perform the necessary modifications for a parallel code
if [ $nb_openmp_processes -gt 0 ] ; then
    nb_processes=$nb_openmp_processes
elif [ $nb_mpi_processes -gt 0 ] ; then
    nb_processes=$nb_mpi_processes
else
    nb_processes=1
fi

sed -e "/^#!\/bin\/ksh.*/d" \
    -e "\%^. /etc/profile > /dev/null 2>&1.*%d" \
    -e "s/time_job_nqs/$time_job_nqs/g" \
    -e "s/memory_job_nqs/$memory_job_nqs/g" \
    -e "s/name_nqs/$name_nqs/g" \
    -e "s/nb_processes_ll/$nb_processes/g" \
    temp/job_tournesol_exec_pid$$ > temp/job_tournesol_exec_pid$$_bis ;

mv temp/job_tournesol_exec_pid$$_bis temp/job_tournesol_exec_pid$$ ;


#_____________________________________________________
#
#       UPLOAD OF THE FILES ON tournesol COMPILATION ON tournesol
#_____________________________________________________
#

printf "\nUpload of the files on tournesol...\n"

# we create a directory on tournesol to copy the executable after the compilation on tournesol.
dir_save_tournesol_upload=$dir_save_tournesol/ns3d_upload_${date_hour}_pid$$
ssh -x tournesol "mkdir -p $dir_save_tournesol_upload"

# we compile in the Scratch directory of tournesol
printf "\nTransfer of the files on tournesol for the compilation...\n"
printf "Directory of compilation: /scratch/deloncle/ns3d_compil_tournesol_pid$$\n"

ssh -x tournesol "mkdir -p /scratch/deloncle/ns3d_compil_tournesol_pid$$"

scp -r     $dir_source/*                       tournesol:/scratch/deloncle/ns3d_compil_tournesol_pid$$
scp        temp/config.h_pid$$                 tournesol:/scratch/deloncle/ns3d_compil_tournesol_pid$$/config.h
scp        temp/Makefile_pid$$                 tournesol:/scratch/deloncle/ns3d_compil_tournesol_pid$$/Makefile


printf "\nCompilation on tournesol...\n"
ssh -x tournesol ". /etc/profile > /dev/null 2>&1 ; cd /scratch/deloncle/ns3d_compil_tournesol_pid$$ ; make ; cp ns3d  $dir_save_tournesol_upload ; cd .. ; rm -rf ns3d_compil_tournesol_pid$$"




#_____________________________________
#
#       UPLOAD OF THE FILES ON tournesol
#_____________________________________
#

printf "\nUpload of the files on tournesol...\n"


echo $dir_source
echo $dir_save_tournesol_upload

# we upload all the necessary files (sources, data file, configuration file) in the working directory
rcp -r     $dir_source/*                       tournesol:$dir_save_tournesol_upload
rcp        temp/config.h_pid$$                 tournesol:$dir_save_tournesol_upload/config.h
rcp        temp/data.in_pid$$                  tournesol:$dir_save_tournesol_upload/data.in
rcp        temp/Makefile_pid$$                 tournesol:$dir_save_tournesol_upload/Makefile


# we upload the input data files if necessary
if [ -f "$velocity_init" ]; then
    printf "\n   Copy of the initial velocity from griffu to tournesol...\n"
    rcp    $velocity_init         tournesol:$dir_save_tournesol_upload/velocity.init
fi

if [ -f "$base2d_init" ]; then
    printf "\n   Copy of the base state from griffu to tournesol...\n"
    rcp    $base2d_init           tournesol:$dir_save_tournesol_upload/base2D.init
fi

if [ $velocity_init_tournesol ]; then
    printf "\n   Copy of the initial velocity from tournesol to tournesol...\n"
    ssh -x tournesol "cp -p $velocity_init_tournesol $dir_save_tournesol_upload/velocity.init"
fi

if [ $base2d_init_tournesol ]; then
    printf "\n   Copy of the initial velocity from tournesol to tournesol...\n"
    ssh -x tournesol "cp -p $base2d_init_tournesol $dir_save_tournesol_upload/base2D.init"
fi



#_______________________________________________
#
#       UPLOAD OF THE EXECUTION SCRIPT ON tournesol
#_______________________________________________
#

printf "\nUpload of the execution script on tournesol...\n"

# we create a directory to save the execution script and we upload it
ssh -x tournesol "mkdir -p ns3d_job_tournesol"
scp temp/job_tournesol_exec_pid$$ "tournesol:ns3d_job_tournesol/job_tournesol_exec_pid$$"

# we delete the useless configuration files
rm temp/config.h_pid$$ temp/data.in_pid$$ temp/job_tournesol_exec_pid$$



#_____________________________________________________
#
#       EXECUTION / SUBMISSION OF THE JOB
#_____________________________________________________
#

# we are in batch mode: we submit the execution script
printf "\nSubmission of the job with PBS...\n"
ssh -x tournesol ". /etc/profile > /dev/null 2>&1 ; cd ns3d_job_tournesol ; chmod 744 job_tournesol_exec_pid$$ ; qsub job_tournesol_exec_pid$$"

printf "\n********** END OF THE NS3D SCRIPT *********\n"

exit
