#!/bin/sh

# I modified this script compared to job_uqbar.sh to avoid uploading a big initial velocity field and to copy it directly from gaya

#
# This script makes easier using the NS3D code on the NEC of IDRIS
# It performs all the necessary steps:
#     - copying the necessary files (input data, sources...) into the execution directory
#     - compilation
#     - execution
#     - save of the results in the save directory
#
# It must be run from the computer connected to the exterior "griffu"

#____________________________________________________________
#
#    PATHS/NAMES
#____________________________________________________________
#

# ------------------------------
# Directories/files used
# ------------------------------

# directory where the results are saved on Gaya (a sub-directory is automatically created for each run)
dir_save_gaya="/u/rech/bll/rbll200/NS3D"

# directory where the source files are
dir_source="../source"

# optional paths to the input data files (previous results data file to restart a run, initial velocity, base flow)
velocity_init=""
base2d_init="/toon/usr/deloncle/base2D_512x512_big_endian"

# name of the simulation, useful to determine the name of the working/save folders
run_name="contrarotating"

# Number of parallel OpenMP processes (0 means non-OpenMP code)
nb_openmp_processes=0

# Mode of execution ("batch" or "interactive")
mode="batch"

# In batch mode, it is necessary to indicate some additional parameters
#   - name of the job (optional)
#   - cpu time in seconds. The unit must not be indicated.
#   - memory used and the unit used (Mb, Gb)
name_nqs="ns3d"
time_cpu_nqs="15000"
memory_nqs="2Gb"


#____________________________________________________________
#
#    COMPILATION PARAMETERS (config.h)
#____________________________________________________________
#

mkdir -p temp
cat > temp/config.h_pid$$ << EOF
! *******************************************
!    NS3D: compilation parameters
! *******************************************

! This file is used by the pre-processor during the compilation step.

! Avoid multiple inclusions of this file. Do not change!
#ifndef CONFIG_H
#define CONFIG_H

! Number of colocation points
#define DIMX 512
#define DIMY 512
#define DIMZ 3

! Padding along each direction
#define PPADKX 0
#define PPADKY 0
#define PPADKZ 0

! The following time schemes are available:
!     - AB2: Adams-Bashforth of order 2
!     - RK2: Runge and Kutta of order 2
!     - RK3: Runge and Kutta of order 3
!     - RK4: Runge and Kutta of order 4
#define RK4

! We can simulate a single complex kz mode for a perturbative and linearized run.
! Warning: for correct results, the eigenmode growthrate must be purely real.
#undef SINGLE_MODE

! we estimate a translation velocity of a basic state
#undef calcul_ut

#endif
EOF



#____________________________________________________________
#
#    SIMULATION PARAMETERS (data.in)
#____________________________________________________________
#

cat > temp/data.in_pid$$ << EOF
********************************************
*    NS3D: simulation parameters           *
********************************************

This file is read at every simulation start.

*** discretization variables ***
lx___________________________   8.4
ly___________________________   8.4
lz___________________________   0.57
dt___________________________   0.005
begin________________________   0.
itmax________________________   15000
de_aliasing__________________   T
    squared:1_or_elliptic:2__   1
    radius_truncation_x______   0.95
    radius_truncation_y______   0.95
    radius_truncation_z______   1.E10

*** physical variables ***
viscosity____________________   0.0003
stratified___________________   T
    brunt_vaisala_frequency__   12.
    schmidt_number___________   1.
2omega_______________________   0.

*** Type of simulation ***
perturbative_________________   T
    linear___________________   T

*** Base state (only perturbation run) ***
base2D_type__________________   file

*** Initial velocity ***
velo_type____________________   null
white_noise__________________   0.01

*** Output ***
output1_period_______________   100
output2_period_______________   100000
output3_period_______________   0


*** additionnal variables... ***

* Stuart vortices *
concentration_stuart_________   0.25

* File gaussian vortices *
file_nb_vortices_____________   2
Vortex 1
    position_x_______________   2.1
    position_y_______________   2.35
    circulation______________   2.
    core_radius______________   0.2
Vortex 2
    position_x_______________   2.1
    position_y_______________   1.85
    circulation______________   -2.
    core_radius______________   0.2

* Random gaussian vortices *
rnd_nb_vortices______________   10
rnd_min_distance_____________   1.
rnd_mean_gamma_______________   6.28
rnd_std_gamma________________   0.
rnd_mean_radius______________   1.
rnd_std_radius_______________   0.

EOF


#____________________________________________________________
#
#     DO NOT MODIFY BELOW!
#____________________________________________________________
#

printf "\n********** START OF THE NS3D SCRIPT *********\n"


#_____________________________________________________
#
#    PRELIMINARY TASKS
#_____________________________________________________
#

# useful to personalize the names of the used directories
date_hour=`date +"%d-%m-%y_%Hh%M"`
pid_job=$$

# we read some simulation parameters in the configuration files
NX=`grep "#define DIMX" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMX/ {print $3}'`
NY=`grep "#define DIMY" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMY/ {print $3}'`
NZ=`grep "#define DIMZ" temp/config.h_pid$$ | head -n 1 | awk '/^#define DIMZ/ {print $3}'`
LX=`awk '/^lx/ {print $2}' temp/data.in_pid$$`
LY=`awk '/^ly/ {print $2}' temp/data.in_pid$$`
LZ=`awk '/^lz/ {print $2}' temp/data.in_pid$$`
Dt=`awk '/^dt/ {print $2}' temp/data.in_pid$$`
nu=`awk '/^viscosity/ {print $2}' temp/data.in_pid$$`
stratifie=`awk '/^stratified/ {print $2}' temp/data.in_pid$$`
brunt=`awk '/brunt_vaisala_frequency/ {print $2}' temp/data.in_pid$$`

if [ "$stratifie" = "F" ]; then
    brunt=0
fi

dir_name="${run_name}_${NX}x${NY}x${NZ}_L=${LX}x${LY}x${LZ}_Dt=${Dt}_nu=${nu}_N=${brunt}"


# we determine the Makefile to use depending on the number of threads
if [ $nb_openmp_processes -eq 0 ]; then
    makefile_file=${dir_source}/Makefile_examples/Makefile_NEC_ASL
else
    makefile_file=${dir_source}/Makefile_examples/Makefile_NEC_ASL_OpenMP
fi

printf "Makefile: ${makefile_file}\n"

#_____________________________________________________
#
#    MODIFICATION OF THE EXECUTION SCRIPT "job_uqbar_exec"
#_____________________________________________________
#

# in batch mode, no redirection of the output messages
if [ "$mode" = "batch" ] ; then
    arg1=""
else
    arg1=$1
fi


sed -e s\%^dir_save_gaya=.*%dir_save_gaya=$dir_save_gaya% \
    -e "s/^arg1=.*/arg1="$1"/" \
    -e "s/^date_hour=.*/date_hour=$date_hour/" \
    -e "s/^pid_job=.*/pid_job=$pid_job/" \
    -e "s/^dir_name=.*/dir_name=$dir_name/" \
    job_uqbar_exec.in > temp/job_uqbar_exec_pid$$

sed "s/-DASL_OPENMP=[0-9]*/-DASL_OPENMP=$nb_openmp_processes/" $makefile_file > temp/Makefile_pid$$

# in batch mode, we write the submission directives in the script header
if [ "$mode" == "batch" ] ; then
    sed -e "/^#!\/bin\/ksh.*/d" \
        -e "\%^. /etc/profile > /dev/null 2>&1.*%d" \
        -e "s/time_cpu_nqs/$time_cpu_nqs/g" \
        -e "s/memory_nqs/$memory_nqs/g" \
        -e "s/name_nqs/$name_nqs/g" \
        temp/job_uqbar_exec_pid$$ > temp/job_uqbar_exec_pid$$_bis ;

    mv temp/job_uqbar_exec_pid$$_bis temp/job_uqbar_exec_pid$$ ;
fi


#_____________________________________
#
#       UPLOAD OF THE FILES ON GAYA
#_____________________________________
#

printf "\nUpload of the files on Gaya...\n"

# we create a directory on gaya to copy all the uploaded files
dir_save_gaya_upload=$dir_save_gaya/ns3d_upload_${date_hour}_pid$$
ssh -x rbll200@gaya.idris.fr "mkdir -p $dir_save_gaya_upload"

# we upload all the necessary files (sources, data file, configuration file) in the working directory
rcp -r     $dir_source/*                       rbll200@gaya.idris.fr:$dir_save_gaya_upload
rcp        temp/config.h_pid$$                 rbll200@gaya.idris.fr:$dir_save_gaya_upload/config.h
rcp        temp/data.in_pid$$                  rbll200@gaya.idris.fr:$dir_save_gaya_upload/data.in
rcp        temp/Makefile_pid$$                 rbll200@gaya.idris.fr:$dir_save_gaya_upload/Makefile


# we upload the input data files if necessary
if [ -f "$velocity_init" ]; then
    rcp -v   $velocity_init         rbll200@gaya.idris.fr:$dir_save_gaya_upload/velocity.init
fi

if [ -f "$base2d_init" ]; then
    rcp -v   $base2d_init           rbll200@gaya.idris.fr:$dir_save_gaya_upload/base2D.init
fi


#_______________________________________________
#
#       UPLOAD OF THE EXECUTION SCRIPT ON UQBAR
#_______________________________________________
#

printf "\nUpload of the execution script on Uqbar...\n"

# we create a directory to save the execution script and we upload it
ssh -x rbll200@uqbar.idris.fr "mkdir -p ns3d_job_uqbar"
rcp temp/job_uqbar_exec_pid$$ "rbll200@uqbar.idris.fr:ns3d_job_uqbar/job_uqbar_exec_pid$$"

# we delete the useless configuration files
rm temp/config.h_pid$$ temp/data.in_pid$$ temp/job_uqbar_exec_pid$$


#_____________________________________________________
#
#       COMPILATION ON RHODES
#_____________________________________________________
#

printf "\nCompilation on Rhodes...\n"

ssh -x rbll200@rhodes.idris.fr "cd $dir_save_gaya_upload ; make"


#_____________________________________________________
#
#       EXECUTION / SUBMISSION OF THE JOB
#_____________________________________________________
#

# we are in batch mode: we submit the execution script
if [ "$mode" = "batch" ] ; then
    printf "\nSubmission of the job with NQS...\n"
    ssh -x rbll200@uqbar.idris.fr "cd ns3d_job_uqbar ; chmod 744 job_uqbar_exec_pid$$ ; qsub job_uqbar_exec_pid$$"

    printf "\n********** END OF THE NS3D SCRIPT *********\n"

# we are in interactive mode: we do NOTHING! The execution script which was uploaded should be executed manually on Uqbar.
elif [ "$mode" = "interactive" ] ; then
    ssh -x rbll200@uqbar.idris.fr "cd ns3d_job_uqbar ; chmod 744 job_uqbar_exec_pid$$"
    printf "\nIt is time to connect on Uqbar and launch the execution script ns3d_job_uqbar/job_uqbar_exec_pid$$\n"

else
    printf "\nERROR : the execution mode (batch or interactive) is incorrect\n"
fi

exit
