program main

    use timing
	use fft

    implicit none

    ! dimensions of the data file
    integer, parameter :: nx=128
    integer, parameter :: ny=128
    integer, parameter :: nz=64
    
    ! translation velocity of the simulation frame
    double precision, parameter :: vx_translation=0.431
    double precision, parameter :: vy_translation=0.
	
	! should we write the dissipation fields?
    logical, parameter :: write_dissipation=.false.
	logical, parameter :: write_R=.false.
	logical, parameter :: write_dudz_dvdz_drhodz=.true.

	!***********************************
	! do not modify from this point
	!***********************************

	! dimensions of the storage array
	integer, parameter :: dx = 2*(nx/2+1)
	integer, parameter :: dy = ny
	integer, parameter :: dz = nz

	! number of modes in spectral space
	integer, parameter :: dkx = dx/2
	integer, parameter :: dky = ny
	integer, parameter :: dkz = nz

    ! maximum number of data files
    integer, parameter :: nb_max_data_file_list=200
    character(len=200), dimension(nb_max_data_file_list) :: tab_data_file_list
	character(len=200) :: data_file_list      
    double precision, dimension(nb_max_data_file_list) :: instants
	integer :: nb_file

    ! data
    double precision, dimension(0:dx-1,0:dy-1,0:dz-1) :: vox, voy, voz, rho, wx, wy, wz
    
	! working arrays useful to compute R, R_tilde and the dissipation
    double precision, dimension(0:dx-1,0:dy-1,0:dz-1) :: wk1, wk2, wk3, wk4,wk5
   
    ! working array useful to read the data
    double precision, dimension(0:nx-1,0:ny-1,0:nz-1) :: wk_io

    ! kinetic (hor, vert, total), potential energy, enstrophy energy (hor, vert, total) and R
    double precision, dimension(nb_max_data_file_list) :: energy_h, energy_v, energy_t, potential_energy,enstrophy_h, &
                                                          enstrophy_v, enstrophy_t, R, R_tilde, dissipation, &
														  enstrophy_stretching, enstrophy_barocline, enstrophy_viscous, &
                                                          dissipation2, diffusion_rho, enstrophy_h_stretching, &
														  enstrophy_h_tilting, enstrophy_h_barocline, enstrophy_h_viscous, &
														  enstrophy_z_stretching, enstrophy_z_tilting, enstrophy_z_viscous

    ! physical dimensions of the box
    double precision :: lx,ly,lz

	! viscosity and Brunt-Vaisala frequency
	double precision :: nu, xns, xnsqr, schm
    
    ! wavenumbers
    double precision, dimension(0:dkx-1) :: kx
    double precision, dimension(0:dky-1) :: ky
    double precision, dimension(0:dkz-1) :: kz

! working array for the FFT. Its size depends on the FFT library.
#if defined(JMFFT)
    double precision, dimension(100 + 2*(nx+ny+nz) + 512*max(nx,ny,nz)) :: fft_work
#elif defined(MATHKEISAN)
    double precision, dimension(2*(2*(nx+ny+nz)+195) + 4*256*max(nx,ny,nz)) :: fft_work
#elif (defined(MATHKEISAN) && DIMZ==1)
    double precision, dimension(2*(2*(nx+ny)+130) + 6*nx*ny) :: fft_work
#elif defined(FFTW_MPI)
    double precision, dimension(2,0:dkx-1,0:dky-1,0:dkz-1,2), save :: fft_work
#else
    double precision, dimension(1), save :: fft_work
#endif
	
	integer :: ix,iy,iz, i_file, eof
	

    write(*,*) '************************************'
    write(*,*) '******    ENERGY ANALYSIS     ******'
    write(*,*) '************************************'

    !***********************
	! We read the list of data file names and we print it to the screen
    !***********************
	write(*,*) 'List of data file names:'
    read(*,*) data_file_list    
	
	nb_file = 0
    open (unit=8,file=data_file_list,form="formatted",status="old",action="read",position="rewind",iostat=eof)   
    do while (eof==0)
        nb_file = nb_file + 1
        read(8,*,iostat=eof)  tab_data_file_list(nb_file)
    end do
    close(8)
    nb_file = nb_file - 1
	
	! we print the names
    do i_file=1, nb_file
        write(*,*) '    ', trim(tab_data_file_list(i_file))
    end do

    write(*,*)
    write(*,*) '   --> Number of data files: ', nb_file

    !***********************
    ! we read the metadata in the first file and we print them
    !***********************
	call timer_start('main')    
	call read_metadata(tab_data_file_list(1))

    write(*,*)
    write(*,*) 'Number of points: ', nx, ny, nz
    write(*,'(A15,G12.5,TR3,G12.5,TR3,G12.5)') 'Dimensions:  ', lx, ly, lz
    write(*,'(A14,G12.5)') 'Viscosity:  ', nu
	write(*,'(A28,G12.5)') 'Brunt Vaisala frequency:  ', xns
    write(*,'(A19,G12.5)') 'Schmidt number:  ', schm
	write(*,*)


    !***********************
    ! we set the FFT and the wavenumbers
    !***********************
	call init_k(kx,ky,kz)

    write(*,*)
    write(*,*) 'Initialization of the Fast Fourier Transformation (FFT)'
    call init_fft(vox,nx,ny,nz,dx,dy,dz,fft_work)
    write(*,*) '  -> ', fft_lib_name



    !***********************
	! we start to analyse each data file
    !***********************

	write(*,*)
	write(*,*) 'Starting analysis...'
	
	do i_file=1,nb_file

        ! we read the instant of the record
        open (unit=8,file=tab_data_file_list(i_file),access='direct',form='unformatted',recl=34)
        read(8,rec=2) instants(i_file)
        close(8)
        
        write(*,*) '   time=',instants(i_file)
    
        ! we read the data
    	call timer_start('reading data')		
    	call read_data(tab_data_file_list(i_file), vox, voy, voz, rho, wx, wy, wz)
    	call timer_stop('reading data')
    
        ! Kinetic energy, potential energy, enstrophy and R
    	call timer_start('computing')	 
    
    	call timer_start('energy')
    	call compute_energy(vox,voy,voz, energy_h(i_file),energy_v(i_file),energy_t(i_file))
    	call timer_stop('energy')
    
    	call timer_start('potential_energy')
    	call compute_potential_energy(rho, potential_energy(i_file))		
    	call timer_stop('potential_energy')
    
    	call timer_start('enstrophy')
    	call compute_enstrophy(wx,wy,wz, enstrophy_h(i_file),enstrophy_v(i_file),enstrophy_t(i_file))       
    	call timer_stop('enstrophy')		
    
    	call timer_start('R')        
    	call compute_R(vox,voy,voz,R(i_file),write_R)  
        call timer_stop('R')
    
    	call timer_start('R_tilde')        
    	call compute_R_tilde(vox,voy,voz,R_tilde(i_file))  
        call timer_stop('R_tilde')
    	
    	call timer_start('dissipation')        
    	call compute_dissipation(vox,voy,voz,dissipation(i_file),write_dissipation)  
        call timer_stop('dissipation')
    
    	call timer_start('dissipation2')        
    	call compute_dissipation2(vox,voy,voz,dissipation2(i_file))  
        call timer_stop('dissipation2')
    
    	call timer_start('diffusion_rho')        
        call compute_diffusion_rho(rho,diffusion_rho(i_file))  
    	call timer_stop('diffusion_rho')
    
    	call timer_start('enstrophy_details')        
    	call compute_enstrophy_details(vox,voy,voz,rho,wx,wy,wz,enstrophy_stretching(i_file),enstrophy_barocline(i_file),&
                                       enstrophy_viscous(i_file),enstrophy_h_stretching(i_file), &
                                       enstrophy_h_tilting(i_file), enstrophy_h_barocline(i_file),&
                                       enstrophy_h_viscous(i_file), enstrophy_z_stretching(i_file), &
                                       enstrophy_z_tilting(i_file),enstrophy_z_viscous(i_file))  
        call timer_stop('enstrophy_details')
    
    	
    	call timer_start('dudz_dvdz_drhodz')        
    	call compute_dudz_dvdz_drhodz(vox,voy,voz,rho,write_dudz_dvdz_drhodz)  
        call timer_stop('dudz_dvdz_drhodz')
    	
    	call timer_stop('computing')
    end do

    open (unit=76,file='energy_analysis_results',form='unformatted')
    write(76) 1,nx,ny,nz,lx,ly,lz,nu,xns,schm
    write(76) nb_file
    write(76) instants(1:nb_file)
    write(76) energy_h(1:nb_file)
    write(76) energy_v(1:nb_file)
    write(76) energy_t(1:nb_file)
    write(76) potential_energy(1:nb_file)
    write(76) enstrophy_h(1:nb_file)
    write(76) enstrophy_v(1:nb_file)
    write(76) enstrophy_t(1:nb_file)
	write(76) R(1:nb_file)
	write(76) R_tilde(1:nb_file)
	write(76) dissipation(1:nb_file)
	write(76) dissipation2(1:nb_file)
	write(76) diffusion_rho(1:nb_file)
    write(76) enstrophy_stretching(1:nb_file)
    write(76) enstrophy_barocline(1:nb_file)
    write(76) enstrophy_viscous(1:nb_file)
    write(76) enstrophy_h_stretching(1:nb_file)
	write(76) enstrophy_h_tilting(1:nb_file)    
	write(76) enstrophy_h_barocline(1:nb_file)
    write(76) enstrophy_h_viscous(1:nb_file)
    write(76) enstrophy_z_stretching(1:nb_file)
	write(76) enstrophy_z_tilting(1:nb_file)    
    write(76) enstrophy_z_viscous(1:nb_file)
    close(76)
	
	call timer_stop('main')
    call timer_report_all

    contains

!**************************
    subroutine compute_enstrophy_details(vox,voy,voz,rho,wox,woy,woz,enstrophy_stretching, &
                                           enstrophy_barocline,enstrophy_viscous, enstrophy_h_stretching, &
                                           enstrophy_h_tilting, enstrophy_h_barocline,enstrophy_h_viscous, &
										   enstrophy_z_stretching, enstrophy_z_tilting,enstrophy_z_viscous)
!**************************
	! computes which terms are important in the enstrophy evolution

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: vox, voy, voz, rho, wox, woy, woz
        double precision, intent(out) :: enstrophy_stretching,enstrophy_barocline,enstrophy_viscous, &
                                         enstrophy_h_stretching,enstrophy_h_tilting, enstrophy_h_barocline, &
										 enstrophy_h_viscous, enstrophy_z_stretching, enstrophy_z_tilting, &
                                         enstrophy_z_viscous

        integer :: ix,iy,iz

		! first term: enstrophy_h stretching
		enstrophy_stretching = 0
		enstrophy_h_stretching = 0
		enstrophy_h_tilting = 0
		enstrophy_z_stretching = 0
		enstrophy_z_tilting = 0
		
		call compute_grad(vox,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					enstrophy_stretching = enstrophy_stretching + wox(ix,iy,iz)* (wox(ix,iy,iz)*wk1(ix,iy,iz)+ &
                                    woy(ix,iy,iz)*wk2(ix,iy,iz)+woz(ix,iy,iz)*wk3(ix,iy,iz))

					enstrophy_h_stretching = enstrophy_h_stretching + &
                                             wox(ix,iy,iz)*wox(ix,iy,iz)*2*wk1(ix,iy,iz) + &
                                             wox(ix,iy,iz)*woy(ix,iy,iz)*wk2(ix,iy,iz) + &
                                             wox(ix,iy,iz)*woz(ix,iy,iz)*wk3(ix,iy,iz) + &
											 woy(ix,iy,iz)*wox(ix,iy,iz)*wk2(ix,iy,iz)

					enstrophy_z_stretching = enstrophy_z_stretching + &
                                             woz(ix,iy,iz)*wox(ix,iy,iz)*wk3(ix,iy,iz)

					enstrophy_h_tilting = enstrophy_h_tilting + &
                                          wox(ix,iy,iz)*woz(ix,iy,iz)*wk3(ix,iy,iz)
                                          										 
          		end do
        	end do
		end do

		call compute_grad(voy,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					enstrophy_stretching = enstrophy_stretching + woy(ix,iy,iz)* (wox(ix,iy,iz)*wk1(ix,iy,iz)+ &
                                    woy(ix,iy,iz)*wk2(ix,iy,iz)+woz(ix,iy,iz)*wk3(ix,iy,iz))

					enstrophy_h_stretching = enstrophy_h_stretching + &
                                             wox(ix,iy,iz)*woy(ix,iy,iz)*wk1(ix,iy,iz) + &
                                             woy(ix,iy,iz)*wox(ix,iy,iz)*wk1(ix,iy,iz) + &
                                             woy(ix,iy,iz)*woy(ix,iy,iz)*2*wk2(ix,iy,iz) + &
											 woy(ix,iy,iz)*woz(ix,iy,iz)*wk3(ix,iy,iz)

					enstrophy_z_stretching = enstrophy_z_stretching + &
                                             woz(ix,iy,iz)*woy(ix,iy,iz)*wk3(ix,iy,iz)


					enstrophy_h_tilting = enstrophy_h_tilting + &
                                          woy(ix,iy,iz)*woz(ix,iy,iz)*wk3(ix,iy,iz)

          		end do
        	end do
		end do

		call compute_grad(voz,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					enstrophy_stretching = enstrophy_stretching + woz(ix,iy,iz)* (wox(ix,iy,iz)*wk1(ix,iy,iz)+ &
                                    woy(ix,iy,iz)*wk2(ix,iy,iz)+woz(ix,iy,iz)*wk3(ix,iy,iz))

					enstrophy_h_stretching = enstrophy_h_stretching + &
                                             wox(ix,iy,iz)*woz(ix,iy,iz)*wk1(ix,iy,iz) + &
                                             woy(ix,iy,iz)*woz(ix,iy,iz)*wk2(ix,iy,iz)

					enstrophy_z_stretching = enstrophy_z_stretching + &
                                             woz(ix,iy,iz)*wox(ix,iy,iz)*wk1(ix,iy,iz) + &
                                             woz(ix,iy,iz)*woy(ix,iy,iz)*wk2(ix,iy,iz) + &
                                             woz(ix,iy,iz)*woz(ix,iy,iz)*2*wk3(ix,iy,iz)

					enstrophy_h_tilting = enstrophy_h_tilting &
                                          - wox(ix,iy,iz)*woz(ix,iy,iz)*wk1(ix,iy,iz) &
                                          - woy(ix,iy,iz)*woz(ix,iy,iz)*wk2(ix,iy,iz)
          		end do
        	end do
		end do
       
		enstrophy_stretching = lx*ly*lz/(nx*ny*nz)*enstrophy_stretching	
		
        enstrophy_h_stretching = lx*ly*lz/(2*nx*ny*nz)*enstrophy_h_stretching
		enstrophy_h_tilting = lx*ly*lz/(2*nx*ny*nz)*enstrophy_h_tilting
		enstrophy_z_stretching = lx*ly*lz/(2*nx*ny*nz)*enstrophy_z_stretching
		enstrophy_z_tilting = -enstrophy_h_tilting		
		
		! second term: enstrophy barocline
		call compute_grad(rho,wk1,wk2,wk3)
		enstrophy_barocline = 0
		do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					enstrophy_barocline = enstrophy_barocline + wox(ix,iy,iz)*wk2(ix,iy,iz)-woy(ix,iy,iz)*wk1(ix,iy,iz)
          		end do
        	end do
		end do
		
		enstrophy_barocline = -xnsqr*lx*ly*lz/(nx*ny*nz)*enstrophy_barocline
		enstrophy_h_barocline = enstrophy_barocline
		

		! third term: enstrophy viscous
		call compute_laplacien(wox,woy,woz,wk1,wk2,wk3)
		enstrophy_viscous = 0
		enstrophy_h_viscous = 0
		enstrophy_z_viscous = 0
		do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					enstrophy_viscous = enstrophy_viscous + wox(ix,iy,iz)*wk1(ix,iy,iz)+woy(ix,iy,iz)*wk2(ix,iy,iz)+&
                                                            woz(ix,iy,iz)*wk3(ix,iy,iz)

					enstrophy_h_viscous = enstrophy_h_viscous + wox(ix,iy,iz)*wk1(ix,iy,iz)+woy(ix,iy,iz)*wk2(ix,iy,iz)
					enstrophy_z_viscous = enstrophy_z_viscous + woz(ix,iy,iz)*wk3(ix,iy,iz)
          		end do
        	end do
		end do

		enstrophy_viscous = lx*ly*lz/(nx*ny*nz)*nu*enstrophy_viscous
		enstrophy_h_viscous = lx*ly*lz/(nx*ny*nz)*nu*enstrophy_h_viscous
		enstrophy_z_viscous = lx*ly*lz/(nx*ny*nz)*nu*enstrophy_z_viscous

end subroutine


!**************************
    subroutine compute_R(vox, voy, voz, R, write_field)
!**************************
	! computes R which is the ratio of the inertial terms and the viscosity term

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: vox, voy, voz
        double precision, intent(out) :: R
		logical, intent(in) :: write_field
        character(len=50) :: file_name
		character(len=10) :: sufx
		integer :: ix,iy,iz

		call compute_grad(vox,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					wk4(ix,iy,iz) = (vox(ix,iy,iz)*wk1(ix,iy,iz) + voy(ix,iy,iz)*wk2(ix,iy,iz) + voz(ix,iy,iz)*wk3(ix,iy,iz))**2
          		end do
        	end do
		end do

		call compute_grad(voy,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					wk4(ix,iy,iz) = wk4(ix,iy,iz) + (vox(ix,iy,iz)*wk1(ix,iy,iz) + voy(ix,iy,iz)*wk2(ix,iy,iz) + &
                                        voz(ix,iy,iz)*wk3(ix,iy,iz))**2
          		end do
        	end do
		end do

		call compute_grad(voz,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					wk4(ix,iy,iz) = wk4(ix,iy,iz) + (vox(ix,iy,iz)*wk1(ix,iy,iz) + voy(ix,iy,iz)*wk2(ix,iy,iz) + &
										voz(ix,iy,iz)*wk3(ix,iy,iz))**2
          		end do
        	end do
		end do
       
		call compute_laplacien(vox,voy,voz,wk1,wk2,wk3)

		do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					wk4(ix,iy,iz) = wk4(ix,iy,iz)/(nu**2*(wk1(ix,iy,iz)**2 + wk2(ix,iy,iz)**2 + wk3(ix,iy,iz)**2))
          		end do
        	end do
		end do
		
		R = 0
		do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					R = R + wk4(ix,iy,iz)
          		end do
        	end do
		end do

		R = sqrt(R/(nx*ny*nz))

		! we write the full dissipation field if necessary
		if (write_field) then
        	call time2string(instants(i_file),sufx)
        	file_name='R.'//sufx
        	open (76,file=file_name,form='unformatted')
        	write(76) 1, nx, ny, nz, lx, ly, lz, nu
			write(76) instants(i_file)
			call write_array(wk4,76)
			close(76)			
		end if


end subroutine

!**************************
    subroutine compute_diffusion_rho(rho, diffusion_rho)
!**************************
	! computes R which is the ratio of the inertial terms and the viscosity term

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: rho
        double precision, intent(out) :: diffusion_rho

        integer :: ix,iy,iz

		call compute_laplacien(rho,wk4,wk5,wk1,wk2,wk3)
        diffusion_rho = 0
		do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					diffusion_rho = diffusion_rho + rho(ix,iy,iz)*wk1(ix,iy,iz)
          		end do
        	end do
		end do

		diffusion_rho = xnsqr*nu/schm*lx*ly*lz/(nx*ny*nz)*diffusion_rho

end subroutine


!**************************
    subroutine compute_dissipation(vox, voy, voz, dissipation, write_field)
!**************************
	! computes the dissipation field

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: vox, voy, voz
        double precision, intent(out) :: dissipation
		logical, intent(in) :: write_field
        character(len=50) :: file_name
		character(len=10) :: sufx
        
		integer :: ix,iy,iz

		! we start with the first term: 2*nu*(du/dx**2+dv/dy**2+dw/dz**2)
		call compute_grad(vox,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					wk5(ix,iy,iz) = 2*nu*wk1(ix,iy,iz)**2
          		end do
        	end do
		end do

		call compute_grad(voy,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					wk5(ix,iy,iz) = wk5(ix,iy,iz) + 2*nu*wk2(ix,iy,iz)**2
          		end do
        	end do
		end do

		call compute_grad(voz,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					wk5(ix,iy,iz) = wk5(ix,iy,iz) + 2*nu*wk3(ix,iy,iz)**2
          		end do
        	end do
		end do
       		
		! we add the second term: nu*[(dv/dz+dw/dy)**2 + (dw/dx+du/dz)**2 + (du/dy+dv/dx)**2]
		call compute_grad(voy,wk1,wk2,wk3)
		wk4=wk3
		call compute_grad(voz,wk1,wk2,wk3)

        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					wk5(ix,iy,iz) = wk5(ix,iy,iz) + nu*(wk4(ix,iy,iz)+wk2(ix,iy,iz))**2
          		end do
        	end do
		end do

		
		call compute_grad(voz,wk1,wk2,wk3)
		wk4=wk1
		call compute_grad(vox,wk1,wk2,wk3)

        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					wk5(ix,iy,iz) = wk5(ix,iy,iz) + nu*(wk4(ix,iy,iz)+wk3(ix,iy,iz))**2
          		end do
        	end do
		end do

		
		call compute_grad(vox,wk1,wk2,wk3)
		wk4=wk2
		call compute_grad(voy,wk1,wk2,wk3)

        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					wk5(ix,iy,iz) = wk5(ix,iy,iz) + nu*(wk4(ix,iy,iz)+wk1(ix,iy,iz))**2
          		end do
        	end do
		end do

		! we compute the total dissipation
        dissipation = 0
		do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					dissipation = dissipation + wk5(ix,iy,iz)
          		end do
        	end do
		end do

		dissipation = lx*ly*lz/(nx*ny*nz)*dissipation

		! we write the full dissipation field if necessary
		if (write_field) then
        	call time2string(instants(i_file),sufx)
        	file_name='dissipation.'//sufx
        	open (76,file=file_name,form='unformatted')
        	write(76) 1, nx, ny, nz, lx, ly, lz, nu
			write(76) instants(i_file)
			call write_array(wk5,76)
			close(76)			
		end if

end subroutine

!**************************
subroutine compute_dudz_dvdz_drhodz(vox, voy, voz, rho, write_field)
!**************************
	    implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: vox, voy, voz, rho
        logical, intent(in) :: write_field
        character(len=50) :: file_name
		character(len=10) :: sufx
        
		integer :: ix,iy,iz

		if (write_field) then		
    		call compute_grad(vox,wk1,wk2,wk3)
            do iz=0, nz-1
    			do iy=0, ny-1
              		do ix=0, nx-1           	
    					wk4(ix,iy,iz) = wk3(ix,iy,iz)
              		end do
            	end do
    		end do
                   
    		call compute_grad(voy,wk1,wk2,wk3)
    		do iz=0, nz-1
    			do iy=0, ny-1
              		do ix=0, nx-1           	
    					wk5(ix,iy,iz) = wk3(ix,iy,iz)
              		end do
            	end do
    		end do

    		call compute_grad(rho,wk1,wk2,wk3)
    		        		    
    		! we write the terms dudz and dvdz		
        	call time2string(instants(i_file),sufx)
        	file_name='dudz_dvdz_drhodz.'//sufx
        	open (76,file=file_name,form='unformatted')
        	write(76) 1, nx, ny, nz, lx, ly, lz, nu, xnsqr
			write(76) instants(i_file)
			call write_array(wk4,76)
			call write_array(wk5,76)
			call write_array(wk3,76)
			close(76)			
    	end if

end subroutine



!**************************
    subroutine compute_dissipation2(vox, voy, voz, dissipation2)
!**************************
	! computes the dissipation field

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: vox, voy, voz
        double precision, intent(out) :: dissipation2
       
		integer :: ix,iy,iz

		dissipation2 = 0
		call compute_laplacien(vox,voy,voz,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					dissipation2 = dissipation2 + vox(ix,iy,iz)*wk1(ix,iy,iz)+voy(ix,iy,iz)*wk2(ix,iy,iz)+voz(ix,iy,iz)*wk3(ix,iy,iz)
          		end do
        	end do
		end do

		dissipation2 = -nu*lx*ly*lz/(nx*ny*nz)*dissipation2

end subroutine


! *************************************************************************
    subroutine time2string(time,string_time)
! *************************************************************************
    !****f* subfunctions/time2string
    !  NAME
    !    time2string
    !
    !  DESCRIPTION
    !    Convert time to a formatted string like "t=0000.000"
    !
    !  INPUTS
    !    time: time to convert in a string
    !
    !  OUTPUT
    !    string_time: string with the desired format
    !***

        implicit none

        double precision, intent(in) :: time
        character(len=10), intent(out) :: string_time

        write(string_time,'(A2,I4.4,A1,I3.3)') 't=',int(time),'.',nint(1000*(time-int(time)))
  end subroutine


! ***********************************************************
    subroutine write_array(pa,id_file)
! ***********************************************************
    !****f* output/write_array
    !  NAME
    !    write_array
    !
    !  DESCRIPTION
    !    Allow to write a subsection of an array into a file without using too much stack memory which
    !    can be an issue on some systems with limited stack memory.
    !    Indeed, surprisingly the following instructions are responsible for the use of an amount of
    !    stack memory corresponding to the **sum** of all the recorded arrays:
    !        write(1) aax(0:nx-1,0:ny-1,0:nz-1)
    !        write(1) aay(0:nx-1,0:ny-1,0:nz-1)
    !        (...)
    !
    !    This subroutine makes use of a single storage array that is static and so defined in data memory.
    !
    !  INPUTS
    !    pa: field to record in physical space
    !    id_file: identifier of the file where the field should be recorded
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1),intent(in) :: pa
        integer, intent(in) :: id_file
        integer :: ix, iy, iz

        do iz=0,nz-1
            do iy=0,ny-1
                do ix=0,nx-1
                    wk_io(ix,iy,iz) = pa(ix,iy,iz)
                end do
            end do
        end do

        write(id_file) wk_io

    end subroutine



!**************************
    subroutine compute_R_tilde(vox, voy, voz, R_tilde)
!**************************
	! computes R_tilde

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: vox, voy, voz
        double precision, intent(out) :: R_tilde
		double precision :: A, B

        integer :: ix,iy,iz
		
		A=0
		call compute_grad(vox,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					A = A + ((vox(ix,iy,iz)+vx_translation)*wk1(ix,iy,iz) + voy(ix,iy,iz)*wk2(ix,iy,iz) + voz(ix,iy,iz)*wk3(ix,iy,iz))**2
          		end do
        	end do
		end do

		call compute_grad(voy,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					A = A + ((vox(ix,iy,iz)+vx_translation)*wk1(ix,iy,iz) + voy(ix,iy,iz)*wk2(ix,iy,iz) + voz(ix,iy,iz)*wk3(ix,iy,iz))**2
          		end do
        	end do
		end do

		call compute_grad(voz,wk1,wk2,wk3)
        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					A = A + ((vox(ix,iy,iz)+vx_translation)*wk1(ix,iy,iz) + voy(ix,iy,iz)*wk2(ix,iy,iz) + voz(ix,iy,iz)*wk3(ix,iy,iz))**2
          		end do
        	end do
		end do
       
		call compute_laplacien(vox,voy,voz,wk1,wk2,wk3)
		
		B=0
		do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					B = B + wk1(ix,iy,iz)**2 + wk2(ix,iy,iz)**2 + wk3(ix,iy,iz)**2
          		end do
        	end do
		end do
		
		R_tilde = sqrt(A/(nu**2*B))

	end subroutine


	subroutine compute_grad(a, dxa, dya, dza)

		implicit none
		
		double precision, dimension(2,0:dx/2-1,0:dy-1,0:dz-1), intent(in) :: a
		double precision, dimension(2,0:dx/2-1,0:dy-1,0:dz-1), intent(out) :: dxa, dya, dza
		double precision :: temp
		integer :: ikx, iky, ikz

		dxa = a
		
		call fwd_fft(dxa, fft_work)
		dya = dxa
		dza = dxa

		do ikz=0,dkz-1
            do iky=0,dky-1
                do ikx=0,dkx-1
					temp = dxa(1,ikx,iky,ikz)
					dxa(1,ikx,iky,ikz) = - kx(ikx)*dxa(2,ikx,iky,ikz)
					dxa(2,ikx,iky,ikz) =   kx(ikx)*temp
				
					temp = dya(1,ikx,iky,ikz)
					dya(1,ikx,iky,ikz) = - ky(iky)*dya(2,ikx,iky,ikz)
					dya(2,ikx,iky,ikz) =   ky(iky)*temp

					temp = dza(1,ikx,iky,ikz)
					dza(1,ikx,iky,ikz) = - kz(ikz)*dza(2,ikx,iky,ikz)
					dza(2,ikx,iky,ikz) =   kz(ikz)*temp
				end do
			end do
		end do

		call bck_fft(dxa, fft_work)
		call bck_fft(dya, fft_work)
		call bck_fft(dza, fft_work)
					
	end subroutine


subroutine compute_laplacien(ax, ay, az, Dax, Day, Daz)

		implicit none
		
		double precision, dimension(2,0:dx/2-1,0:dy-1,0:dz-1), intent(in) :: ax, ay, az
		double precision, dimension(2,0:dx/2-1,0:dy-1,0:dz-1), intent(out) :: Dax, Day, Daz
		integer :: ikx, iky, ikz
		double precision :: k2

		Dax = ax
		Day = ay
		Daz = az
		
		call fwd_fft(Dax, fft_work)
		call fwd_fft(Day, fft_work)
		call fwd_fft(Daz, fft_work)

		do ikz=0,dkz-1
            do iky=0,dky-1
                do ikx=0,dkx-1
					k2 = kx(ikx)**2+ky(iky)**2+kz(ikz)**2
					Dax(:,ikx,iky,ikz) = -k2*Dax(:,ikx,iky,ikz)
					Day(:,ikx,iky,ikz) = -k2*Day(:,ikx,iky,ikz)
					Daz(:,ikx,iky,ikz) = -k2*Daz(:,ikx,iky,ikz)					
				end do
			end do
		end do

		call bck_fft(Dax, fft_work)
		call bck_fft(Day, fft_work)
		call bck_fft(Daz, fft_work)
					
	end subroutine


!**************************
    subroutine compute_energy(vox, voy, voz, energy_h, energy_v, energy_t)
!**************************

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: vox, voy, voz
        double precision, intent(out) :: energy_h, energy_v, energy_t

        integer :: ix,iy,iz

        energy_h = 0
        energy_v = 0
        energy_t = 0       

        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					energy_h = energy_h + (vox(ix,iy,iz)+vx_translation)**2 + (voy(ix,iy,iz)+vy_translation)**2
					energy_v = energy_v + voz(ix,iy,iz)**2
          		end do
        	end do
		end do

        energy_h = lx*ly*lz/(2*nx*ny*nz)*energy_h
        energy_v = lx*ly*lz/(2*nx*ny*nz)*energy_v

        energy_t = energy_h + energy_v

end subroutine


!**************************
    subroutine compute_enstrophy(wx, wy, wz, enstrophy_h, enstrophy_v, enstrophy_t)
!**************************

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: wx, wy, wz
        double precision, intent(out) :: enstrophy_h, enstrophy_v, enstrophy_t

        integer :: ix,iy,iz

        enstrophy_h = 0
        enstrophy_v = 0
        enstrophy_t = 0       

        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					enstrophy_h = enstrophy_h + wx(ix,iy,iz)**2 + wy(ix,iy,iz)**2
					enstrophy_v = enstrophy_v + wz(ix,iy,iz)**2
          		end do
        	end do
		end do

        enstrophy_h = lx*ly*lz/(2*nx*ny*nz)*enstrophy_h
        enstrophy_v = lx*ly*lz/(2*nx*ny*nz)*enstrophy_v

        enstrophy_t = enstrophy_h + enstrophy_v

end subroutine


!**************************
    subroutine compute_potential_energy(rho, potential_energy)
!**************************

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: rho
        double precision, intent(out) :: potential_energy

        integer :: ix,iy,iz

        potential_energy = 0

        do iz=0, nz-1
			do iy=0, ny-1
          		do ix=0, nx-1           	
					potential_energy = potential_energy + rho(ix,iy,iz)**2
          		end do
        	end do
		end do

        potential_energy = xnsqr*lx*ly*lz/(2*nx*ny*nz)*potential_energy

end subroutine



!**************************
    subroutine read_metadata(data_file_name)
!**************************
!    we read the metadata and we check the dimensions in the file

        implicit none

        character(len=200), intent(in) :: data_file_name
        integer :: flag, nxread, nyread, nzread
		double precision :: trash_double
		integer :: trash_int

        ! we check the dimensions
        open (unit=8,file=data_file_name,form='unformatted')
        read(8) flag,nxread,nyread,nzread,lx,ly,lz, trash_double, trash_int, trash_int, trash_double, &
                trash_double, trash_double, nu, trash_int, xns, schm
        close(8)
        if (flag /= 1 .or. nxread /= nx .or. nyread /= ny .or. nzread /= nz) then
            write(*,*) "ERROR: the data file has bad format. Failure."
            stop
        end if

		xnsqr = xns**2

    end subroutine

!**************************
    subroutine read_data(data_file_name,vox, voy, voz, rho, wx, wy, wz)
!**************************

        implicit none

        character(len=200), intent(in) :: data_file_name
        double precision, dimension(0:dx-1,0:dy-1,0:dz-1),intent(out) :: vox, voy, voz, rho, wx, wy, wz
        double precision :: trash_double
		integer, dimension(31) :: trash_header

		open (unit=8,file=data_file_name,form='unformatted')
		read(8) trash_header
		read(8) trash_double
		call read_array(vox,8)
		call read_array(voy,8)
		call read_array(voz,8)
		call read_array(rho,8)
		call read_array(wx,8)
		call read_array(wy,8)
		call read_array(wz,8)
		close(8)

    end subroutine

! ***********************************************************
    subroutine read_array(pa,id_file)
! ***********************************************************

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1),intent(out) :: pa
        integer, intent(in) :: id_file
        integer :: ix, iy, iz

        read(id_file) wk_io

        do iz=0,nz-1
            do iy=0,ny-1
                do ix=0,nx-1
                    pa(ix,iy,iz) = wk_io(ix,iy,iz)
                end do
            end do
        end do

    end subroutine



! **************************************************************************
    subroutine init_k(kx,ky,kz)
! **************************************************************************
    !****f* subfunctions/init_k
    !  NAME
    !    init_k
    !
    !  DESCRIPTION
    !    Define the wavenumbers
    !
    !  OUTPUT
    !    kx,ky,kz: the wavenumbers
    !***

        implicit none

        double precision, dimension(0:dkx-1), intent(out) :: kx
        double precision, dimension(0:dky-1), intent(out) :: ky
        double precision, dimension(0:dkz-1), intent(out) :: kz
        double precision :: pi
        integer :: ik

        pi = acos(-1.)

        ! Only the positive modes are stored along the x direction.
        ! The negative modes are the complex conjugates of the positive ones.
        do ik=0,dkx-1
            kx(ik)=ik*2.*pi/lx
        end do

        do ik=0,dky-1
            if (ik <= dky/2) then
                ky(ik)=ik*2.*pi/ly
            else
                ky(ik)=(ik-dky)*2.*pi/ly
            end if
        end do

        do ik=0,dkz-1
            if (ik <= dkz/2) then
                kz(ik)=ik*2.*pi/lz
            else
                kz(ik)=(ik-dkz)*2.*pi/lz
            end if
        end do

    end subroutine

end program
