!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organixation : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************

!***************************************************************************
module timing
!***************************************************************************
!****h* ns3d/timing
!  NAME
!    timing
!
!  DESCRIPTION
!    Provide subroutines to easily time and profile a code. It
!    allows the user to get and print the cpu-time spent in sections of
!    a code: main program, subroutines, nested subroutines etc.
!
!  EXAMPLE
!    Here is a simple example of how to use the timing module:
!
!        program main
!
!            implicit none
!
!            use timing                     ! we call the timing module
!            call timer_start("timer1")     ! we launch a timer
!
!            (...)                          ! section of the code we want to profile
!
!            call timer_stop("timer1")      ! we stop the timer
!            call timer_report_all          ! we print all the timing results
!
!        end program
!***

    implicit none

    private

#ifdef MPI
    include "mpif.h"
    integer, save :: code, rank
#endif

    integer, parameter :: nb_timer_max = 100

    type timer
        character(len=22) :: name
        integer :: id_parent
        integer :: depth
        integer :: iter
        double precision :: counter_start
        double precision :: counter_sum
    end type

    type(timer), dimension(nb_timer_max) :: tab_timer
    integer :: nb_timer=0, id_last=0, current_depth=0
    logical, save :: timer_active = .true.

    public :: timer_start, timer_stop, timer_report_tree, timer_report_flat, timer_report_all, timer_set_active

    contains


!***************************************************************************
    subroutine timer_start(name, nb_iter)
!***************************************************************************
    !****f* timing/timer_start
    !  NAME
    !    timer_start
    !
    !  DESCRIPTION
    !    Create a new timer or relaunch it if an identical timer (same name and depth)
    !    already exists.
    !
    !  INPUTS
    !    name: name of the timer to create or relaunch
    !    nb_iter (optional): number added to the "#calls" counter at every call of timer_start
    !                        (default is 1). This is useful when a single timer includes in fact
    !                        several call to the same subroutine and that we want to count exactly
    !                        the calls to that subroutine.
    !
    !  NOTES
    !    The timer is not created if the static module variable "timer_active" is set to .false.
    !***

        implicit none

        character(len=*), intent(in) :: name
        integer, optional, intent(in) :: nb_iter
        integer :: id_existing_timer, iter_step
        integer :: i

        if (.not. timer_active) then
            return
        end if

        ! if necessary, we set a new iter_step to modify how the calls are countered
        if (present(nb_iter)) then
            iter_step = nb_iter
        else
            iter_step = 1
        end if

        current_depth = current_depth + 1

        ! we search for a timer with the same name and the same parent
        id_existing_timer = 0
        do i=1, nb_timer
            if (tab_timer(i)%name == name .and. tab_timer(i)%id_parent == id_last) then
                id_existing_timer = i
            end if
        end do

        ! we create a new timer
        if (id_existing_timer==0) then
            nb_timer = nb_timer + 1

            tab_timer(nb_timer)%name = name
            tab_timer(nb_timer)%id_parent = id_last
            tab_timer(nb_timer)%depth = current_depth
            tab_timer(nb_timer)%iter = iter_step
            tab_timer(nb_timer)%counter_start = timing_cpu()
            tab_timer(nb_timer)%counter_sum = 0

            id_last = nb_timer

        ! we update an existing timer
        else
            tab_timer(id_existing_timer)%id_parent = id_last
            tab_timer(id_existing_timer)%depth = current_depth
            tab_timer(id_existing_timer)%iter = tab_timer(id_existing_timer)%iter + iter_step
            tab_timer(id_existing_timer)%counter_start = timing_cpu()

            id_last = id_existing_timer
        end if

    end subroutine

!***************************************************************************
    subroutine timer_stop(name)
!***************************************************************************
    !****f* timing/timer_stop
    !  NAME
    !    timer_stop
    !
    !  DESCRIPTION
    !    Stop an existing timer.
    !
    !  INPUTS
    !    name: name of the timer to stop
    !***

        implicit none

        character(len=*), intent(in) :: name

        if (.not. timer_active) then
            return
        end if

        ! we check if the name is consistent with the name of the last timer
        if (tab_timer(id_last)%name /= name) then
            write(*,*) 'Timer error : the results will be wrong'
        end if

        current_depth = current_depth - 1
        tab_timer(id_last)%counter_sum = tab_timer(id_last)%counter_sum + timing_cpu() - tab_timer(id_last)%counter_start
        id_last = tab_timer(id_last)%id_parent

    end subroutine


!***************************************************************************
    subroutine timer_report_tree
!***************************************************************************
    !****f* timing/timer_report_tree
    !  NAME
    !    timer_report_tree
    !
    !  DESCRIPTION
    !    Print a full report of the timers in a tree: it shows the number of calls and the cpu
    !    time spent in each timer and its nested timers. It also indicate the percentage of time
    !    spent in every timer compared to its parent timer.
    !
    !  EXAMPLE
    !    Here is a short example of the output created by timer_report_tree:
    !
    !        -------------------------------------------------------------------------------
    !        TIMING REPORT: TREE
    !        name                     #calls            sec             % time
    !        -------------------------------------------------------------------------------
    !         main                         1           22.43             100.00
    !        -------------------------------------------------------------------------------
    !         initialization               1            1.37               6.11
    !         time_stepping                1           20.95              93.40
    !           time_scheme               98             19.06              90.97
    !             non_linear_term         98               14.97              78.51
    !               curl                  98                  0.67               4.50
    !               fft                 1274                  9.99              66.76
    !               vect_prod             98                  0.71               4.78
    !               urho                  98                  0.43               2.89
    !               special_div           98                  0.36               2.40
    !               #others                -                  2.80              18.69
    !             #others                  -                4.10              21.49
    !           projection                98              0.39               1.88
    !           de_aliasing                98              0.21               1.01
    !           test_nan                  98              0.17               0.83
    !           output2                   10              1.11               5.31
    !           output1                    1              0.00               0.01
    !         #others                      -            0.11               0.49
    !        -------------------------------------------------------------------------------
    !***

        implicit none

        character (len=40), parameter :: spaces = ' '
        double precision, dimension(nb_timer_max) :: counter_sum_level
        double precision :: percent_sum, counter_sum_parent
        integer :: id_temp, id_parent_temp
        integer :: indent
        integer :: i, j

        counter_sum_level = 0

        write(*,'(a79)')'----------------------------------------------------------------------------------'
        write(*,'(a)')'TIMING REPORT: TREE'
        write(*,'(a22,a9,a15,a19)') 'name'//spaces,'#calls','sec','% time'

        do i=1,nb_timer

            counter_sum_level(tab_timer(i)%depth) = counter_sum_level(tab_timer(i)%depth) + tab_timer(i)%counter_sum

            indent=max(1,tab_timer(i)%depth*2-3)

            ! upper level (no parent)
            if (tab_timer(i)%depth == 1) then
                write(*,'(a79)')'----------------------------------------------------------------------------------'
                percent_sum = 100
            else
                percent_sum = 100*tab_timer(i)%counter_sum/max(tab_timer(tab_timer(i)%id_parent)%counter_sum,1.D-100)
            end if

            write(*,'(a22,i9,a,f15.2,f19.2)')                              &
                repeat(' ',indent)//tab_timer(i)%name,                     &
                tab_timer(i)%iter,                                         &
                repeat(' ',indent),                                        &
                tab_timer(i)%counter_sum,                                  &
                percent_sum

            ! upper level (no parent)
            if (tab_timer(i)%depth == 1) then
                write(*,'(a79)')'----------------------------------------------------------------------------------'
            end if

            ! we report the missing part for that depth
            id_parent_temp = i
            do j=1,tab_timer(i)%depth-tab_timer(i+1)%depth

                id_temp = id_parent_temp
                id_parent_temp = tab_timer(id_temp)%id_parent
                counter_sum_parent = tab_timer(id_parent_temp)%counter_sum
                indent=max(1,tab_timer(id_temp)%depth*2-3)

                if (counter_sum_parent-counter_sum_level(tab_timer(id_temp)%depth) > 0.005) then
                    write(*,'(a22,a9,a,f15.2,f19.2)')                                                    &
                        repeat(' ',indent)//'#others'//spaces,                                      &
                        '-', &
                        repeat(' ',indent),                                        &
                        counter_sum_parent-counter_sum_level(tab_timer(id_temp)%depth),             &
                        100*(counter_sum_parent-counter_sum_level(tab_timer(id_temp)%depth))/max(counter_sum_parent,1.D-100)
                end if

                counter_sum_level(tab_timer(id_temp)%depth) = 0
             end do

        end do

        write(*,'(a79)')'----------------------------------------------------------------------------------'

    end subroutine


!***************************************************************************
    subroutine timer_report_flat
!***************************************************************************
    !****f* timing/timer_report_flat
    !  NAME
    !    timer_report_flat
    !
    !  DESCRIPTION
    !    Print a full report of the timers in a flat way. The timers that have the same name but different
    !    depths are summed up. It shows the number of calls, the total time spent in a timer and the
    !    self time spent only in a timer and not its nested timers. It also indicates the percentage of
    !    self time compared to the time elapsed in the upper level timers (usually the whole program).
    !
    !  EXAMPLE
    !    Here is a short example of the output created by timer_report_flat:
    !
    !        -------------------------------------------------------------------------------
    !        TIMING REPORT: FLAT
    !        name                     #calls       total sec        self sec        % time
    !        -------------------------------------------------------------------------------
    !         fft                       1274            9.99            9.99         44.53
    !         time_scheme                 98           19.06            4.10         18.26
    !         non_linear_term             98           14.97            2.80         12.47
    !         time_stepping                1           20.95            1.89          8.44
    !         initialization               1            1.37            1.37          6.11
    !         output2                     10            1.11            1.11          4.96
    !         vect_prod                   98            0.71            0.71          3.19
    !         curl                        98            0.67            0.67          3.00
    !         urho                        98            0.43            0.43          1.93
    !         projection                  98            0.39            0.39          1.75
    !         special_div                 98            0.36            0.36          1.60
    !         de_aliasing                  98            0.21            0.21          0.94
    !         test_nan                    98            0.17            0.17          0.77
    !         **main                       1           22.43            0.11          0.49
    !         output1                      1            0.00            0.00          0.01
    !        -------------------------------------------------------------------------------
    !***

        implicit none

        type timer_reported
            character(len=22) :: name
            integer :: iter
            double precision :: counter_sum
            double precision :: counter_self_sum
        end type

        type(timer_reported), dimension(nb_timer_max) :: tab_timer_sorted
        logical, dimension(nb_timer_max) :: timer_done
        character (len=40), parameter :: spaces = ' '
        double precision :: total_time
        integer :: nb_timer_reported
        character (len=22) :: temp_name
        integer :: temp_iter
        double precision :: temp_counter_sum
        double precision :: temp_counter_self_sum
        integer :: i, j

        timer_done = .false.
        nb_timer_reported = 0
        total_time = 0

        do i=1, nb_timer

            ! this a new timer that has not already been reported
            if (.not. timer_done(i)) then
                timer_done(i) = .true.

                temp_name = tab_timer(i)%name
                temp_iter = tab_timer(i)%iter
                temp_counter_sum = tab_timer(i)%counter_sum
                temp_counter_self_sum = self_time(i)

                ! if this is an upper timer, we update the total time
                if (tab_timer(i)%depth == 1) then
                    total_time = total_time + tab_timer(i)%counter_sum
                    temp_name = '**'//temp_name
                end if

                ! we search forward for timers with the same name
                do j=i+1, nb_timer
                    if (tab_timer(i)%name == tab_timer(j)%name) then
                        timer_done(j) = .true.
                        temp_iter = temp_iter + tab_timer(j)%iter
                        temp_counter_sum = temp_counter_sum + tab_timer(j)%counter_sum
                        temp_counter_self_sum = temp_counter_self_sum + self_time(j)
                    end if
                end do

                ! we sort the new timer into the reported_timer_array
                j = nb_timer_reported
                do while (j>=1 .and. tab_timer_sorted(j)%counter_self_sum < temp_counter_self_sum)
                    tab_timer_sorted(j+1) = tab_timer_sorted(j)
                    j = j-1
                end do

                tab_timer_sorted(j+1)%name = temp_name
                tab_timer_sorted(j+1)%iter = temp_iter
                tab_timer_sorted(j+1)%counter_sum = temp_counter_sum
                tab_timer_sorted(j+1)%counter_self_sum = temp_counter_self_sum

                nb_timer_reported = nb_timer_reported + 1
            end if
        end do

        ! we print the results which are sorted
        write(*,'(a79)')'----------------------------------------------------------------------------------'
        write(*,'(a)')'TIMING REPORT: FLAT'
        write(*,'(a22,a9,a16,a16,a14)') 'name'//spaces,'#calls','total sec','self sec','% time'
        write(*,'(a79)')'----------------------------------------------------------------------------------'


        do i=1, nb_timer_reported
            write(*,'(a22,i9,f16.2,f16.2,f14.2)')                               &
                ' '//tab_timer_sorted(i)%name,                                  &
                tab_timer_sorted(i)%iter,                                       &
                tab_timer_sorted(i)%counter_sum,                                &
                tab_timer_sorted(i)%counter_self_sum,                           &
                100*tab_timer_sorted(i)%counter_self_sum/total_time

        end do

        write(*,'(a79)')'----------------------------------------------------------------------------------'

    end subroutine

#ifdef MPI
!***************************************************************************
    subroutine timer_report_all
!***************************************************************************
    !****f* timing/timer_report_all
    !  NAME
    !    timer_report_all
    !
    !  DESCRIPTION
    !    Print both the tree report and the flat report.
    !***

        implicit none

        call MPI_comm_rank(MPI_COMM_WORLD,rank,code)

        if (rank==0) then
            write(*,'(a79)')'----------------------------------------------------------------------------------'
            write(*,'(a)')'MPI CODE: ONLY ONE PROCESS (RANK 0) IS SHOWN BELOW'
            write(*,'(a79)')'----------------------------------------------------------------------------------'

            call timer_report_tree

            write(*,*)

            call timer_report_flat
        end if

    end subroutine


#else
!***************************************************************************
    subroutine timer_report_all
!***************************************************************************
    !****f* timing/timer_report_all
    !  NAME
    !    timer_report_all
    !
    !  DESCRIPTION
    !    Print both the tree report and the flat report.
    !***

        implicit none

        call timer_report_tree

        write(*,*)

        call timer_report_flat

    end subroutine
#endif

!***************************************************************************
    double precision function self_time(id_timer)
!***************************************************************************
    !****f* timing/self_time
    !  NAME
    !    self_time
    !
    !  DESCRIPTION
    !    Return the self-cpu-time elapsed in a timer and not in its nested timers.
    !
    !  INPUTS
    !    id_timer: id of the timer whose self cpu time is required
    !
    !  OUTPUT
    !    self_time: self time spent in seconds
    !***

        implicit none

        integer, intent(in) :: id_timer
        integer :: i

        ! we determine the self-time by substracting the time of the children timers
        i = id_timer+1
        self_time = tab_timer(id_timer)%counter_sum

        do while (i<= nb_timer .and. tab_timer(i)%id_parent == id_timer)
            self_time = self_time - tab_timer(i)%counter_sum
            i = i + 1
        end do

    end function


!***************************************************************************
    double precision function timing_cpu()
!***************************************************************************
    !****f* timing/timing_cpu
    !  NAME
    !    timing_cpu
    !
    !  DESCRIPTION
    !    Return the cpu time in seconds.
    !
    !  OUTPUT
    !    timing_cpu: cpu-time in seconds
    !***

        implicit none

        call cpu_time(timing_cpu)

    end function


!***************************************************************************
    subroutine timer_set_active(timer_status)
!***************************************************************************
    !****f* timing/timer_set_active
    !  NAME
    !    timer_set_active
    !
    !  DESCRIPTION
    !    Allow to switch between active (default) and inactive mode. In inactive mode, new timers are
    !    not created when calling the subroutine timer_start. It is a way of not timing some parts
    !    of the code.
    !
    !  INPUTS
    !    timer_status:  new timer status (.false. or .true.)
    !***

        implicit none

        logical, intent(in) :: timer_status

        timer_active = timer_status

    end subroutine

end module timing
