!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organixation : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************

#include "config.h"
#include "extended_config.h"

! *************************************************************************
module subfunctions
! *************************************************************************
!****h* ns3d/subfunctions
!  NAME
!    subfunctions
!
!  DESCRIPTION
!    Contain different functions of various purposes that are used in ns3d. The most important
!    is the one that computes the non-linear term useful for time integration.
!***

    use timing
    use fft
    use global_vars

    implicit none

    contains


! *************************************************************************
    subroutine trunc(svx,svy,svz,srho)
! *************************************************************************
    !****f* subfunctions/precal_exp_fact
    !  NAME
    !    precal_exp_fact
    !
    !  DESCRIPTION
    !    Classical desaliasing in spectral space
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields in spectral space
    !
    !  OUTPUT
    !    svx,svy,svz,srho: velocity and density fields in spectral space with truncated modes
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) ::  svx, svy, svz, srho
        integer :: ik1, ik2, ik3

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    if(.not. trunctab(ik1,ik2,ik3)) then
                        svx(:,ik1,ik2,ik3)  = 0.
                        svy(:,ik1,ik2,ik3)  = 0.
                        svz(:,ik1,ik2,ik3)  = 0.
                        srho(:,ik1,ik2,ik3) = 0.
                    end if
                end do
            end do
        end do

    end subroutine


! *************************************************************************
    subroutine init_trunctab(trunctab)
! *************************************************************************
    !****f* subfunctions/init_trunctab
    !  NAME
    !    init_trunctab
    !
    !  DESCRIPTION
    !    Initialize the de_aliasing table
    !
    !  OUTPUT
    !    trunctab: de_aliasing table that indicates which modes are to be truncated
    !***

        implicit none

        logical, dimension(0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: trunctab
        double precision :: dks
        double precision :: pi
        integer :: ik1, ik2, ik3
        double precision :: r_kxmax2, r_kymax2, r_kzmax2

        pi = acos(-1.)

        r_kxmax2 = kxmax2*rtrunc_x**2
        r_kymax2 = kymax2*rtrunc_y**2
        r_kzmax2 = kzmax2*rtrunc_z**2

        trunctab = .false.
        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1

                    dks=0.

                    if (type_trunc==1) then
                        if (gnkx /= 1) then
                            dks = max(dks,kx2(IKX)/r_kxmax2)
                        end if

                        if (gnky /= 1) then
                            dks = max(dks,ky2(IKY)/r_kymax2)
                        end if

                        if (gnkz /= 1) then
                            dks = max(dks,kz2(IKZ)/r_kzmax2)
                        end if
                    else
                        if (gnkx /= 1) then
                            dks = dks + kx2(IKX)/r_kxmax2
                        end if

                        if (gnky /= 1) then
                            dks = dks + ky2(IKY)/r_kymax2
                        end if

                        if (gnkz /= 1) then
                            dks = dks + kz2(IKZ)/r_kzmax2
                        end if
                    end if

                    ! we add "epsilo" to avoid truncating modes because of rounding errors
                    if (dks <= 1. + epsilo) then
                        trunctab(ik1,ik2,ik3)=.true.
                    end if
                end do
            end do
        end do

    end subroutine


! **************************************************************************
    subroutine init_k(kx,kx2,ky,ky2,kz,kz2,kxmax,kxmax2,kymax,kymax2,kzmax,kzmax2)
! **************************************************************************
    !****f* subfunctions/init_k
    !  NAME
    !    init_k
    !
    !  DESCRIPTION
    !    Define the wavenumbers
    !
    !  OUTPUT
    !    kx,kx2,ky,ky2,kz,kz2: the wavenumbers
    !    cax,ca2x: alternative definition of the x wavenumbers
    !***

        implicit none

        double precision, dimension(0:dkx-1), intent(out) :: kx, kx2
        double precision, dimension(0:dky-1), intent(out) :: ky, ky2
        double precision, dimension(0:dkz-1), intent(out) :: kz, kz2
        double precision, intent(out) :: kxmax, kymax, kzmax, kxmax2, kymax2, kzmax2
        double precision :: pi
        integer :: ik

        pi = acos(-1.)

        ! Only the positive modes are stored along the x direction.
        ! The negative modes are the complex conjugates of the positive ones.
        do ik=0,gnkx-1
            kx(ik)=ik*2.*pi/lx
        end do


! Only a portion of ky modes are stored locally on each processor in MPI
#ifdef MPI
        do ik=0,nky-1
            ! the global iky index (between 0 and gnky-1) is rank+nb_proc*ik and corresponds to the local
            ! index ik of the process rank
            if (rank+nb_proc*ik <= gnky/2) then
                ky(ik)=(rank+nb_proc*ik)*2.*pi/ly
            else
                ky(ik)=(rank+nb_proc*ik-gnky)*2.*pi/ly
            end if
        end do
#else
        do ik=0,gnky-1
            if (ik <= gnky/2) then
                ky(ik)=ik*2.*pi/ly
            else
                ky(ik)=(ik-gnky)*2.*pi/ly
            end if
        end do
#endif


#ifdef SINGLE_MODE
        if (lz==0.) then
            kz = 0.
        else
            kz = 2.*pi/lz
        end if
#else
        do ik=0,gnkz-1
            if (ik <= gnkz/2) then
                kz(ik)=ik*2.*pi/lz
            else
                kz(ik)=(ik-gnkz)*2.*pi/lz
            end if
        end do
#endif

        kx2=kx*kx
        ky2=ky*ky
        kz2=kz*kz

        kxmax=(gnkx-1)*2.*pi/lx
        kymax=gnky/2*2.*pi/ly
        kzmax=gnkz/2*2.*pi/lz

        kxmax2 = kxmax**2
        kymax2 = kymax**2
        kzmax2 = kzmax**2

    end subroutine


! **************************************************************************
    subroutine projection(svx,svy,svz)
! **************************************************************************
    !****f* subfunctions/projection
    !  NAME
    !    projection
    !
    !  DESCRIPTION
    !    Project a velocity field on the plane orthogonal to the wave vector k in spectral space
    !    It insures the incompressibilty of the flow.
    !
    !  INPUTS
    !    svx,svy,svz: velocity field in spectral space
    !
    !  OUTPUT
    !    svx,svy,svz: projected velocity field in spectral space
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) :: svx, svy, svz
        double precision inv_k2, s1, s2
        integer :: ik1, ik2, ik3

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                        ! use of epsilo to avoid null k module
                        inv_k2=1./max(kx2(IKX)+ky2(IKY)+kz2(IKZ),epsilo)

#ifdef SINGLE_MODE
                        s1 =(kx(IKX)*svx(1,ik1,ik2,ik3)+ky(IKY)*svy(1,ik1,ik2,ik3)-kz(IKZ)*svz(2,ik1,ik2,ik3))*inv_k2
                        s2 =(kx(IKX)*svx(2,ik1,ik2,ik3)+ky(IKY)*svy(2,ik1,ik2,ik3)+kz(IKZ)*svz(1,ik1,ik2,ik3))*inv_k2

                        svx(1,ik1,ik2,ik3)= svx(1,ik1,ik2,ik3)-s1*kx(IKX)
                        svx(2,ik1,ik2,ik3)= svx(2,ik1,ik2,ik3)-s2*kx(IKX)
                        svy(1,ik1,ik2,ik3)= svy(1,ik1,ik2,ik3)-s1*ky(IKY)
                        svy(2,ik1,ik2,ik3)= svy(2,ik1,ik2,ik3)-s2*ky(IKY)
                        svz(1,ik1,ik2,ik3)= svz(1,ik1,ik2,ik3)-s2*kz(IKZ)
                        svz(2,ik1,ik2,ik3)= svz(2,ik1,ik2,ik3)+s1*kz(IKZ)
#else
                        s1 =(kx(IKX)*svx(1,ik1,ik2,ik3)+ky(IKY)*svy(1,ik1,ik2,ik3)+kz(IKZ)*svz(1,ik1,ik2,ik3))*inv_k2
                        s2 =(kx(IKX)*svx(2,ik1,ik2,ik3)+ky(IKY)*svy(2,ik1,ik2,ik3)+kz(IKZ)*svz(2,ik1,ik2,ik3))*inv_k2

                        svx(1,ik1,ik2,ik3)= svx(1,ik1,ik2,ik3)-s1*kx(IKX)
                        svx(2,ik1,ik2,ik3)= svx(2,ik1,ik2,ik3)-s2*kx(IKX)
                        svy(1,ik1,ik2,ik3)= svy(1,ik1,ik2,ik3)-s1*ky(IKY)
                        svy(2,ik1,ik2,ik3)= svy(2,ik1,ik2,ik3)-s2*ky(IKY)
                        svz(1,ik1,ik2,ik3)= svz(1,ik1,ik2,ik3)-s1*kz(IKZ)
                        svz(2,ik1,ik2,ik3)= svz(2,ik1,ik2,ik3)-s2*kz(IKZ)
#endif

                end do
            end do
        end do

    end subroutine


! ***************************************************************************
    subroutine init_coord(xx,yy,zz)
! ***************************************************************************
    !****f* subfunctions/init_coord
    !  NAME
    !    init_coord
    !
    !  DESCRIPTION
    !    Initialize the coordinates of the colocation points
    !
    !  OUTPUT
    !    xx,yy,zz: coordinates of the colocation points
    !***

        implicit none

        double precision, dimension(0:dx-1), intent(out) :: xx
        double precision, dimension(0:dy-1), intent(out) :: yy
        double precision, dimension(0:dz-1), intent(out) :: zz
        integer :: i

        do i=0,nx-1
            xx(i)=i*lx/nx
        end do

        do i=0,ny-1
            yy(i)=i*ly/ny
        end do

! Only a slice of the points along the z direction are stored locally in MPI
#ifdef MPI
        do i=0,nz-1
            zz(i)=(rank*nz+i)*lz/gnz
        end do
#else
        do i=0,nz-1
            zz(i)=i*lz/nz
        end do
#endif

    end subroutine


#ifdef TEST_NAN
! ***************************************************************************
    logical function detect_nan(aa)
! ***************************************************************************
    !****f* subfunctions/detect_nan
    !  NAME
    !    detect_nan
    !
    !  DESCRIPTION
    !    Test the existence of "Not a Number" (NaN) in a field. Stops the simulation if it has failed.
    !
    !  INPUTS
    !    aa: the field to test
    !
    !  OUTPUT
    !    detect_nan: logical that indicates if a NaN was found (.true.) or not (.false.)
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: aa
        integer :: ik1, ik2, ik3

        detect_nan = .false.
        do ik3=0, nk3-1
            do ik2=0, nk2-1
                do ik1=0, nk1-1
                    if (isnan(aa(1,ik1,ik2,ik3))) then
                        detect_nan = .true.
                    end if
                end do
            end do
        end do

    end function
#endif


! *************************************************************************
    subroutine curl(svx,svy,svz,sbx,sby,sbz)
! *************************************************************************
    !****f* subfunctions/curl
    !  NAME
    !    curl
    !
    !  DESCRIPTION
    !    Curl (vorticity) of a velocity field in spectral space.
    !
    !  INPUTS
    !    svx,svy,svz: velocity field in spectral space
    !
    !  OUTPUT
    !    sbx,sby,sbz: curl (vorticity) of the input field in spectral space
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: svx, svy, svz
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: sbx, sby, sbz
        integer :: ik1, ik2, ik3

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1

#ifndef SINGLE_MODE
                    sbx(1,ik1,ik2,ik3)=kz(IKZ)*svy(2,ik1,ik2,ik3)-ky(IKY)*svz(2,ik1,ik2,ik3)
                    sbx(2,ik1,ik2,ik3)=ky(IKY)*svz(1,ik1,ik2,ik3)-kz(IKZ)*svy(1,ik1,ik2,ik3)
                    sby(1,ik1,ik2,ik3)=kx(IKX)*svz(2,ik1,ik2,ik3)-kz(IKZ)*svx(2,ik1,ik2,ik3)
                    sby(2,ik1,ik2,ik3)=kz(IKZ)*svx(1,ik1,ik2,ik3)-kx(IKX)*svz(1,ik1,ik2,ik3)
#else
                    sbx(1,ik1,ik2,ik3)=-kz(IKZ)*svy(1,ik1,ik2,ik3)-ky(IKY)*svz(2,ik1,ik2,ik3)
                    sbx(2,ik1,ik2,ik3)= ky(IKY)*svz(1,ik1,ik2,ik3)-kz(IKZ)*svy(2,ik1,ik2,ik3)
                    sby(1,ik1,ik2,ik3)= kx(IKX)*svz(2,ik1,ik2,ik3)+kz(IKZ)*svx(1,ik1,ik2,ik3)
                    sby(2,ik1,ik2,ik3)= kz(IKZ)*svx(2,ik1,ik2,ik3)-kx(IKX)*svz(1,ik1,ik2,ik3)
#endif

                    sbz(1,ik1,ik2,ik3)=ky(IKY)*svx(2,ik1,ik2,ik3)-kx(IKX)*svy(2,ik1,ik2,ik3)
                    sbz(2,ik1,ik2,ik3)=kx(IKX)*svy(1,ik1,ik2,ik3)-ky(IKY)*svx(1,ik1,ik2,ik3)
                 end do
            end do
        end do

    end subroutine


! *************************************************************************
    subroutine vect_prod(pax,pay,paz,pbx,pby,pbz)
! *************************************************************************
    !****f* subfunctions/vect_prod
    !  NAME
    !    vect_prod
    !
    !  DESCRIPTION
    !    Vectorial product of (pax,pay,paz) and (pbx,pby,pbz).
    !    Take into account the terms corresponding to the base flow (linear run) and the Coriolis force
    !
    !  INPUTS
    !    pax,pay,paz: first field in physical space
    !    pbx,pby,pbz: second field in physical space
    !
    !  OUTPUT
    !    pbx,pby,pbz: vectorial product of the two input fields in physical space
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: pax, pay, paz
        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(inout) :: pbx, pby, pbz
        double precision :: vbx0,vby0,vbz0,wbx0,wby0,wbz0
        double precision :: ax0,ay0,az0,bx0,by0,bz0
        integer :: ix, iy, iz

        ! non-perturbative
        if (.not. perturbative) then
            do iz=0,nz-1
                do iy=0,ny-1
                    do ix=0,nx-1
                        ax0=pax(ix,iy,iz)
                        ay0=pay(ix,iy,iz)
                        az0=paz(ix,iy,iz)
                        bx0=pbx(ix,iy,iz)
                        by0=pby(ix,iy,iz)
                        bz0=pbz(ix,iy,iz)

                        pbx(ix,iy,iz)=ay0*(omega2+bz0)-az0*by0
                        pby(ix,iy,iz)=az0*bx0-ax0*(omega2+bz0)
                        pbz(ix,iy,iz)=ax0*by0-ay0*bx0
                    end do
                end do
            end do

        ! perturbative and non-linear
        else if (perturbative .and. (.not. linear)) then
            do iz=0,nz-1
                do iy=0,ny-1
                    do ix=0,nx-1
                        vbx0=vbx(ix,iy)
                        vby0=vby(ix,iy)
                        vbz0=vbz(ix,iy)
                        wbx0=wbx(ix,iy)
                        wby0=wby(ix,iy)
                        wbz0=wbz(ix,iy)+omega2

                        ax0=pax(ix,iy,iz)
                        ay0=pay(ix,iy,iz)
                        az0=paz(ix,iy,iz)
                        bx0=pbx(ix,iy,iz)
                        by0=pby(ix,iy,iz)
                        bz0=pbz(ix,iy,iz)

                        pbx(ix,iy,iz)=vby0*bz0-vbz0*by0 + ay0*(wbz0+bz0) - az0*(wby0+by0)
                        pby(ix,iy,iz)=vbz0*bx0-vbx0*bz0 + az0*(wbx0+bx0) - ax0*(wbz0+bz0)
                        pbz(ix,iy,iz)=vbx0*by0-vby0*bx0 + ax0*(wby0+by0) - ay0*(wbx0+bx0)
                    end do
                end do
            end do

       ! perturbative and linear
        else if (perturbative .and. linear) then
            do iz=0,nz-1
                do iy=0,ny-1
                    do ix=0,nx-1
                        vbx0=vbx(ix,iy)
                        vby0=vby(ix,iy)
                        vbz0=vbz(ix,iy)
                        wbx0=wbx(ix,iy)
                        wby0=wby(ix,iy)
                        wbz0=wbz(ix,iy)+omega2

                        ax0=pax(ix,iy,iz)
                        ay0=pay(ix,iy,iz)
                        az0=paz(ix,iy,iz)
                        bx0=pbx(ix,iy,iz)
                        by0=pby(ix,iy,iz)
                        bz0=pbz(ix,iy,iz)

                        pbx(ix,iy,iz)=vby0*bz0-vbz0*by0 + ay0*wbz0-az0*wby0
                        pby(ix,iy,iz)=vbz0*bx0-vbx0*bz0 + az0*wbx0-ax0*wbz0
                        pbz(ix,iy,iz)=vbx0*by0-vby0*bx0 + ax0*wby0-ay0*wbx0
                    end do
                end do
            end do
        end if

    end subroutine


! *************************************************************************
    subroutine urho(pvx,pvy,pvz,prho)
! *************************************************************************
    !****f* subfunctions/urho
    !  NAME
    !    urho
    !
    !  DESCRIPTION
    !    calculus of (pvx,pvy,pvz).prho in physical space. This is useful for the density equation.
    !
    !  INPUTS
    !    pvx,pvy,pvz,prho: velocity and density fields in physical space
    !
    !  OUTPUT
    !    pvx,pvy,pvz: (pvx,pvy,pvz).prho in physical space
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(inout) :: pvx, pvy, pvz
        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(in) :: prho
        integer :: ix, iy, iz

        ! non-perturbative
        if (.not. perturbative) then
            do iz=0,nz-1
                do iy=0,ny-1
                    do ix=0,nx-1
                        pvx(ix,iy,iz)=prho(ix,iy,iz)*pvx(ix,iy,iz)
                        pvy(ix,iy,iz)=prho(ix,iy,iz)*pvy(ix,iy,iz)
                        pvz(ix,iy,iz)=prho(ix,iy,iz)*pvz(ix,iy,iz)
                    end do
                end do
            end do

        ! perturbative and non-linear
        else if (perturbative .and. (.not. linear)) then
            do iz=0,nz-1
                 do iy=0,ny-1
                    do ix=0,nx-1
                        pvx(ix,iy,iz)=prho(ix,iy,iz)*(pvx(ix,iy,iz)+vbx(ix,iy))
                        pvy(ix,iy,iz)=prho(ix,iy,iz)*(pvy(ix,iy,iz)+vby(ix,iy))
                        pvz(ix,iy,iz)=prho(ix,iy,iz)*(pvz(ix,iy,iz)+vbz(ix,iy))
                    end do
                end do
            end do

       ! perturbative and linear
        else if (perturbative .and. linear) then
            do iz=0,nz-1
                do iy=0,ny-1
                    do ix=0,nx-1
                        pvx(ix,iy,iz)=prho(ix,iy,iz)*vbx(ix,iy)
                        pvy(ix,iy,iz)=prho(ix,iy,iz)*vby(ix,iy)
                        pvz(ix,iy,iz)=prho(ix,iy,iz)*vbz(ix,iy)
                    end do
                end do
            end do
        end if

     end subroutine


! *************************************************************************
    subroutine special_div(sax,say,saz,sbz)
! *************************************************************************
    !****f* subfunctions/special_div
    !  NAME
    !    special_div
    !
    !  DESCRIPTION
    !    calculus of the term sbz - div(sax,say,saz) which is required in the density equation
    !    we directly store the result of the divergence in the sbz field to save one storage field
    !
    !  INPUTS
    !    sax,say,saz: field in spectral space
    !    sbz: field in spectral space
    !
    !  OUTPUT
    !    sbz: output field sbz - div(sax,say,saz) in spectral space
    !***

        implicit none

        double precision ,dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: sax, say, saz
        double precision ,dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) :: sbz
        integer :: ik1, ik2, ik3

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    sbz(1,ik1,ik2,ik3) = sbz(1,ik1,ik2,ik3) + (kx(IKX)*sax(2,ik1,ik2,ik3)+ky(IKY)* &
                                         say(2,ik1,ik2,ik3) + kz(IKZ)*saz(2,ik1,ik2,ik3))
                    sbz(2,ik1,ik2,ik3) = sbz(2,ik1,ik2,ik3) - (kx(IKX)*sax(1,ik1,ik2,ik3)+ky(IKY)* &
                                         say(1,ik1,ik2,ik3) + kz(IKZ)*saz(1,ik1,ik2,ik3))
                end do
            end do
        end do

    end subroutine

! *************************************************************************
    subroutine time2string(time,string_time)
! *************************************************************************
    !****f* subfunctions/time2string
    !  NAME
    !    time2string
    !
    !  DESCRIPTION
    !    Convert time to a formatted string like "t=0000.000"
    !
    !  INPUTS
    !    time: time to convert in a string
    !
    !  OUTPUT
    !    string_time: string with the desired format
    !***

        implicit none

        double precision, intent(in) :: time
        character(len=10), intent(out) :: string_time

        write(string_time,'(A2,I4.4,A1,I3.3)') 't=',int(time),'.',nint(1000*(time-int(time)))
  end subroutine


! *************************************************************************
    subroutine write_text(text,output_format)
! *************************************************************************
    !****f* subfunctions/write_text
    !  NAME
    !    write_text
    !
    !  DESCRIPTION
    !    Write the string 'text' on the screen. Useful to avoid multiple outputs when using MPI parallelization.
    !
    !  INPUTS
    !    text: character string to write
    !    output_format: format to use to print the string 'text'
    !***

        implicit none

        character(len=*), optional, intent(in) :: text
        character(len=*), optional, intent(in) :: output_format

#ifdef MPI
        if (rank==0) then
            if (present(text)) then
                if (present(output_format)) then
                    write(*,output_format) text
                else
                    write(*,*) text
                end if
            else
                write(*,*)
            end if
        end if
#else
        if (present(text)) then
            if (present(output_format)) then
                write(*,output_format) text
            else
                write(*,*) text
            end if
        else
            write(*,*)
        end if
#endif

    end subroutine


! *************************************************************************
    double precision function gaussian(mean,std)
! *************************************************************************
    !****f* subfunctions/gaussian
    !  NAME
    !    gaussian
    !
    !  DESCRIPTION
    !    generate a random number that follows a gaussian (==normal) law of given mean and standard deviation
    !    we use a reject method.
    !
    !  INPUTS
    !    mean: mean of the normal distribution
    !    std: standard deviation of the normal distribution
    !
    !  OUTPUT
    !    gaussian: random number that follows the expected normal law
    !***

        double precision, intent(in) :: mean, std
        double precision :: x, u
        logical :: test
        double precision :: pi

        pi = acos(-1.)
        test = .false.

        do while (.not. test)

            ! uniform random variable between -10.  and 10.
            call random_number(x)
            x = 20.*x-10.

            ! uniform random variable between -5 and +5
            call random_number(u)

            if (u < 1./sqrt(2*pi)*exp(-0.5*x**2)) then
                test = .true.
            end if
        end do

        gaussian = std*x + mean

    end function


! *************************************************************************
    subroutine non_linear_term(svx,svy,svz,srho,sfx,sfy,sfz,sfrho)
! *************************************************************************
    !****f* subfunctions/non_linear_term
    !  NAME
    !    non_linear_term
    !
    !  DESCRIPTION
    !    Evaluation of the non-linear term in spectral space.
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields in spectral space
    !
    !  OUTPUT
    !    svx,svy,svz,srho: velocity and density fields in physical space
    !    sfx,sfy,sfz,sfrho: non-linear term for the velocity and density equations in spectral space
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) :: svx, svy, svz, srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: sfx, sfy, sfz, sfrho

#ifdef calcul_ut
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1) :: sdwz_dx, sdwz_dy
        integer :: ik1, ik2, ik3
        double precision :: ut, error
#endif

        ! save of svz in the field sfrho: trick to save one storage field later
        if (stratification) then
            sfrho=svz
        end if

        ! we determine the vorticity of the velocity field sv
        call timer_start("curl")
        call curl(svx,svy,svz,sfx,sfy,sfz)
        call timer_stop("curl")

#ifdef calcul_ut
        ! we determine the derivatives of wz
        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    sdwz_dx(1,ik1,ik2,ik3) = -kx(IKX)*sfz(2,ik1,ik2,ik3)
                    sdwz_dx(2,ik1,ik2,ik3) = kx(IKX)*sfz(1,ik1,ik2,ik3)

                    sdwz_dy(1,ik1,ik2,ik3) = -ky(IKY)*sfz(2,ik1,ik2,ik3)
                    sdwz_dy(2,ik1,ik2,ik3) = ky(IKY)*sfz(1,ik1,ik2,ik3)
                end do
            end do
        end do
#endif

        ! we switch to physical space
        call timer_start("fft",6)
        call bck_fft(svx,fft_work)
        call bck_fft(svy,fft_work)
        call bck_fft(svz,fft_work)
        call bck_fft(sfx,fft_work)
        call bck_fft(sfy,fft_work)
        call bck_fft(sfz,fft_work)
        call timer_stop("fft")

        if (stratification) then
            call timer_start("fft")
            call bck_fft(srho,fft_work)
            call timer_stop("fft")
        end if

#ifdef calcul_ut
        ! we switch in physical space the derivatives of wz
        call bck_fft(sdwz_dx,fft_work)
        call bck_fft(sdwz_dy,fft_work)
#endif


        ! calculus of V x W in physical space (this is a "pseudo-spectral" method)
        call timer_start("vect_prod")
        call vect_prod(svx,svy,svz,sfx,sfy,sfz)
        call timer_stop("vect_prod")

#ifdef calcul_ut
	! we determine an estimation of u0, the translation speed
	ut = sum((svx(:,0:nk1-1,0:nk2-1,0:nk3-1)*sdwz_dx(:,0:nk1-1,0:nk2-1,0:nk3-1)+svy(:,0:nk1-1,0:nk2-1,0:nk3-1)* &
		sdwz_dy(:,0:nk1-1,0:nk2-1,0:nk3-1))*sdwz_dx(:,0:nk1-1,0:nk2-1,0:nk3-1))

	ut = ut/max(sum(sdwz_dx(:,0:nk1-1,0:nk2-1,0:nk3-1)**2), epsilo)

	! we estimate the estimation error
	error = sum(((svx(:,0:nk1-1,0:nk2-1,0:nk3-1)-ut)*sdwz_dx(:,0:nk1-1,0:nk2-1,0:nk3-1) + &
			svy(:,0:nk1-1,0:nk2-1,0:nk3-1)*sdwz_dy(:,0:nk1-1,0:nk2-1,0:nk3-1))**2)/(nx*ny*nz)

	! we print the results
	write(*,*) '    Estimation of ut:', ut
	write(*,*) '    Estimation of the error on ut:', error
#endif

! calculus of the non-linear term for the density equation
        if (stratification) then

            ! u*rho in physical space
            call timer_start("urho")
            call urho(svx,svy,svz,srho)
            call timer_stop("urho")

            ! switching to spectral space
            call timer_start("fft",3)
            call fwd_fft(svx,fft_work)
            call fwd_fft(svy,fft_work)
            call fwd_fft(svz,fft_work)
            call timer_stop("fft")

            ! u*grad(srho) is calculated by taking the divergence of u*srho because div(u) = 0
            call timer_start("special_div")
            call special_div(svx,svy,svz,sfrho)
            call timer_stop("special_div")

            ! we add the buoyancy term for the velocity equation
            sfz=sfz-xnsqr*srho
        end if

        ! switching to spectral space
        call timer_start("fft",3)
        call fwd_fft(sfx,fft_work)
        call fwd_fft(sfy,fft_work)
        call fwd_fft(sfz,fft_work)
        call timer_stop("fft")

        ! we project the non-linear term to insure the incompressibility of the flow
        call timer_start("projection")
        call projection(sfx,sfy,sfz)
        call timer_stop("projection")

    end subroutine

end module
