!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organixation : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************

!***************************************************************************
module timing
!***************************************************************************
!****h* ns3d/timing
!  NAME
!    timing
!
!  DESCRIPTION
!    Provide subroutines to easily time and profile a code. It
!    allows the user to get and print the cpu-time spent in sections of
!    a code: main program, subroutines, nested subroutines etc.
!
!  EXAMPLE
!    Here is a simple example of how to use the timing module:
!
!        program main
!
!            implicit none
!
!            use timing                     ! we call the timing module
!            call timer_start("timer1")     ! we launch a timer
!
!            (...)                          ! section of the code we want to profile
!
!            call timer_stop("timer1")      ! we stop the timer
!            call timer_report_all          ! we print all the timing results
!
!        end program
!***

    implicit none

    private

#ifdef MPI
    include "mpif.h"
    integer, save :: code, rank
#endif

    integer, parameter :: nb_timer_max = 100
    integer :: nb_periods_max, nb_periods_sec

    type timer
        character(len=15) :: name
        integer :: id_parent
        integer :: depth
        integer :: iter
        double precision :: counter_start_cpu
        double precision :: counter_sum_cpu
        double precision :: counter_sum_elapsed
        integer :: period_start_elapsed
    end type

    type(timer), dimension(nb_timer_max) :: tab_timer
    integer :: nb_timer=0, id_last=0, current_depth=0
    logical, save :: timer_active = .true.
    logical, save :: nb_periods_initialized = .false.

    public :: timer_start, timer_stop, timer_report_tree, timer_report_flat, timer_report_all, timer_set_active

    contains


!***************************************************************************
    subroutine timer_start(name, nb_iter)
!***************************************************************************
    !****f* timing/timer_start
    !  NAME
    !    timer_start
    !
    !  DESCRIPTION
    !    Create a new timer or relaunch it if an identical timer (same name and depth)
    !    already exists.
    !
    !  INPUTS
    !    name: name of the timer to create or relaunch
    !    nb_iter (optional): number added to the "#calls" counter at every call of timer_start
    !                        (default is 1). This is useful when a single timer includes in fact
    !                        several call to the same subroutine and that we want to count exactly
    !                        the calls to that subroutine.
    !
    !  NOTES
    !    The timer is not created if the static module variable "timer_active" is set to .false.
    !***

        implicit none

        character(len=*), intent(in) :: name
        integer, optional, intent(in) :: nb_iter
        integer :: id_existing_timer, iter_step
        integer :: i

        if (.not. timer_active) then
            return
        end if

        ! if necessary, we set a new iter_step to modify how the calls are countered
        if (present(nb_iter)) then
            iter_step = nb_iter
        else
            iter_step = 1
        end if

        current_depth = current_depth + 1

        ! we search for a timer with the same name and the same parent
        id_existing_timer = 0
        do i=1, nb_timer
            if (tab_timer(i)%name == name .and. tab_timer(i)%id_parent == id_last) then
                id_existing_timer = i
            end if
        end do

        ! we create a new timer
        if (id_existing_timer==0) then
            nb_timer = nb_timer + 1

            tab_timer(nb_timer)%name = name
            tab_timer(nb_timer)%id_parent = id_last
            tab_timer(nb_timer)%depth = current_depth
            tab_timer(nb_timer)%iter = iter_step
            tab_timer(nb_timer)%counter_start_cpu = timing_cpu()
            tab_timer(nb_timer)%counter_sum_cpu = 0
            tab_timer(nb_timer)%period_start_elapsed = timing_elapsed()
            tab_timer(nb_timer)%counter_sum_elapsed = 0

            id_last = nb_timer

        ! we update an existing timer
        else
            tab_timer(id_existing_timer)%id_parent = id_last
            tab_timer(id_existing_timer)%depth = current_depth
            tab_timer(id_existing_timer)%iter = tab_timer(id_existing_timer)%iter + iter_step
            tab_timer(id_existing_timer)%counter_start_cpu = timing_cpu()
            tab_timer(id_existing_timer)%period_start_elapsed = timing_elapsed()

            id_last = id_existing_timer
        end if

    end subroutine

!***************************************************************************
    subroutine timer_stop(name)
!***************************************************************************
    !****f* timing/timer_stop
    !  NAME
    !    timer_stop
    !
    !  DESCRIPTION
    !    Stop an existing timer.
    !
    !  INPUTS
    !    name: name of the timer to stop
    !***

        implicit none

        character(len=*), intent(in) :: name
        integer :: nb_periods

        if (.not. timer_active) then
            return
        end if

        if (.not. nb_periods_initialized) then
            call system_clock(count_rate=nb_periods_sec, count_max=nb_periods_max)
            nb_periods_initialized = .true.
        end if


        ! we check if the name is consistent with the name of the last timer
        if (tab_timer(id_last)%name /= name) then
            write(*,*) 'Timer error : the results will be wrong'
        end if

        current_depth = current_depth - 1

        ! cpu time
        tab_timer(id_last)%counter_sum_cpu = tab_timer(id_last)%counter_sum_cpu + timing_cpu() - &
                                             tab_timer(id_last)%counter_start_cpu

        ! elapsed time. We can only count elapsed time that is smaller than nb_periods_max/nb_periods_sec
        nb_periods = timing_elapsed() - tab_timer(id_last)%period_start_elapsed

        if (nb_periods>=0) then
            tab_timer(id_last)%counter_sum_elapsed = tab_timer(id_last)%counter_sum_elapsed + dble(nb_periods)/nb_periods_sec
        else
            tab_timer(id_last)%counter_sum_elapsed = tab_timer(id_last)%counter_sum_elapsed + &
                                                     dble(nb_periods+nb_periods_max)/nb_periods_sec
        end if

        id_last = tab_timer(id_last)%id_parent

    end subroutine


!***************************************************************************
    subroutine timer_report_tree
!***************************************************************************
    !****f* timing/timer_report_tree
    !  NAME
    !    timer_report_tree
    !
    !  DESCRIPTION
    !    Print a full report of the timers in a tree: it shows the number of calls and the cpu
    !    time spent in each timer and its nested timers. It also indicate the percentage of time
    !    spent in every timer compared to its parent timer.
    !
    !  EXAMPLE
    !    Here is a short example of the output created by timer_report_tree:
    !
    !        -------------------------------------------------------------------------------
    !        TIMING REPORT: TREE (in sec)
    !         name             #calls          cpu        elapsed      %cpu   %elapsed
    !        -------------------------------------------------------------------------------
    !         main                  1        41.84         62.08      100.0     100.0
    !        -------------------------------------------------------------------------------
    !         initialization        1        27.97         38.13       66.9      61.4
    !         time_stepping         2        12.54         12.54       30.0      20.2
    !          time_scheme          2         12.41         12.41       99.0      99.0
    !           nonlin_term         8          11.35         11.35       91.4      91.4
    !            curl               8            0.44          0.44        3.9       3.8
    !            fft              104            9.24          9.25       81.5      81.5
    !            vect_prod          8            0.51          0.51        4.5       4.5
    !            urho               8            0.36          0.36        3.2       3.2
    !            special_div        8            0.25          0.25        2.2       2.2
    !            projection         8            0.27          0.27        2.4       2.4
    !            #others            -            0.27          0.28        2.4       2.4
    !           #others             -           1.06          1.07        8.6       8.6
    !          projection           2          0.07          0.07        0.5       0.5
    !          de_aliasing          2          0.01          0.01        0.1       0.1
    !          output1              2          0.03          0.03        0.2       0.2
    !          test_nan             2          0.02          0.02        0.2       0.2
    !         #others               -         1.33         11.41        3.2      18.4
    !        -------------------------------------------------------------------------------
    !***

        implicit none

        character (len=40), parameter :: spaces = ' '
        double precision, dimension(nb_timer_max) :: counter_sum_cpu_level, counter_sum_elapsed_level
        double precision :: percent_sum_cpu, counter_sum_cpu_parent, percent_sum_elapsed, counter_sum_elapsed_parent
        integer :: id_temp, id_parent_temp
        integer :: indent
        integer :: i, j

        counter_sum_cpu_level = 0
        counter_sum_elapsed_level = 0

        write(*,'(a79)')'----------------------------------------------------------------------------------'
        write(*,'(a)')'TIMING REPORT: TREE (in sec)'
        write(*,'(a18,a6,a13,a15,a10,a11)') ' name'//spaces,'#calls','cpu','elapsed','%cpu','%elapsed'

        do i=1,nb_timer

            counter_sum_cpu_level(tab_timer(i)%depth) = counter_sum_cpu_level(tab_timer(i)%depth) + tab_timer(i)%counter_sum_cpu
            counter_sum_elapsed_level(tab_timer(i)%depth) = counter_sum_elapsed_level(tab_timer(i)%depth) + &
                                                            tab_timer(i)%counter_sum_elapsed

            indent=max(1,tab_timer(i)%depth-1)

            ! upper level (no parent)
            if (tab_timer(i)%depth == 1) then
                write(*,'(a79)')'----------------------------------------------------------------------------------'
                percent_sum_cpu = 100
                percent_sum_elapsed = 100
            else
                percent_sum_cpu = 100*tab_timer(i)%counter_sum_cpu/max(tab_timer(tab_timer(i)%id_parent)%counter_sum_cpu,1.D-100)
                percent_sum_elapsed = 100*tab_timer(i)%counter_sum_elapsed / &
                                      max(tab_timer(tab_timer(i)%id_parent)%counter_sum_elapsed,1.D-100)
            end if

            write(*,'(a16,i8,a,f12.2,f14.2,f11.1,f10.1)')                  &
                repeat(' ',indent)//tab_timer(i)%name,                     &
                tab_timer(i)%iter,                                         &
                repeat(' ',indent),                                        &
                tab_timer(i)%counter_sum_cpu,                              &
                tab_timer(i)%counter_sum_elapsed,                          &
                percent_sum_cpu,                                           &
                percent_sum_elapsed


            ! upper level (no parent)
            if (tab_timer(i)%depth == 1) then
                write(*,'(a79)')'----------------------------------------------------------------------------------'
            end if

            ! we report the missing part for that depth
            id_parent_temp = i
            do j=1,tab_timer(i)%depth-tab_timer(i+1)%depth

                id_temp = id_parent_temp
                id_parent_temp = tab_timer(id_temp)%id_parent
                counter_sum_cpu_parent = tab_timer(id_parent_temp)%counter_sum_cpu
                counter_sum_elapsed_parent = tab_timer(id_parent_temp)%counter_sum_elapsed
                indent=max(1,tab_timer(id_temp)%depth-1)

                if (counter_sum_cpu_parent-counter_sum_cpu_level(tab_timer(id_temp)%depth) > 0.005) then
                    write(*,'(a16,a8,a,f12.2,f14.2,f11.1,f10.1)')    &
                        repeat(' ',indent)//'#others'//spaces,    &
                        '-', &
                        repeat(' ',indent),                       &
                        counter_sum_cpu_parent-counter_sum_cpu_level(tab_timer(id_temp)%depth),           &
                        counter_sum_elapsed_parent-counter_sum_elapsed_level(tab_timer(id_temp)%depth),   &
                        100*(counter_sum_cpu_parent-counter_sum_cpu_level(tab_timer(id_temp)%depth)) /    &
                            max(counter_sum_cpu_parent,1.D-100), &
                        100*(counter_sum_elapsed_parent-counter_sum_elapsed_level(tab_timer(id_temp)%depth)) / &
                            max(counter_sum_elapsed_parent,1.D-100)

                end if

                counter_sum_cpu_level(tab_timer(id_temp)%depth) = 0
                counter_sum_elapsed_level(tab_timer(id_temp)%depth) = 0
             end do

        end do

        write(*,'(a79)')'----------------------------------------------------------------------------------'

    end subroutine


!***************************************************************************
    subroutine timer_report_flat
!***************************************************************************
    !****f* timing/timer_report_flat
    !  NAME
    !    timer_report_flat
    !
    !  DESCRIPTION
    !    Print a full report of the timers in a flat way. The timers that have the same name but different
    !    depths are summed up. It shows the number of calls, the total time spent in a timer and the
    !    self time spent only in a timer and not its nested timers. It also indicates the percentage of
    !    self time compared to the time elapsed in the upper level timers (usually the whole program).
    !
    !  EXAMPLE
    !    Here is a short example of the output created by timer_report_flat:
    !
    !        -------------------------------------------------------------------------------
    !        TIMING REPORT: FLAT (in sec)
    !        Time spent in the subroutine but NOT in the nested subroutines
    !        -------------------------------------------------------------------------------
    !        name              #calls             cpu          elapsed     %cpu   %elapsed
    !        -------------------------------------------------------------------------------
    !         initialization        1           27.97           38.13      66.9      61.4
    !         fft                 104            9.24            9.25      22.1      14.9
    !         **main                1            1.33           11.41       3.2      18.4
    !         time_scheme           2            1.06            1.07       2.5       1.7
    !         vect_prod             8            0.51            0.51       1.2       0.8
    !         curl                  8            0.44            0.44       1.0       0.7
    !         urho                  8            0.36            0.36       0.9       0.6
    !         projection           10            0.34            0.34       0.8       0.5
    !         nonlin_term           8            0.27            0.28       0.7       0.4
    !         special_div           8            0.25            0.25       0.6       0.4
    !         time_stepping         2            0.13            0.12       0.3       0.2
    !         output1               2            0.03            0.03       0.1       0.0
    !         test_nan              2            0.02            0.02       0.0       0.0
    !         de_aliasing           2            0.01            0.01       0.0       0.0
    !***     -------------------------------------------------------------------------------

        implicit none

        type timer_reported
            character(len=22) :: name
            integer :: iter
            double precision :: counter_self_sum_cpu
            double precision :: counter_self_sum_elapsed
        end type

        type(timer_reported), dimension(nb_timer_max) :: tab_timer_sorted
        logical, dimension(nb_timer_max) :: timer_done
        character (len=40), parameter :: spaces = ' '
        double precision :: total_time_cpu, total_time_elapsed
        integer :: nb_timer_reported
        character (len=22) :: temp_name
        integer :: temp_iter
        double precision :: temp_counter_self_sum_cpu
        double precision :: temp_counter_self_sum_elapsed
        integer :: i, j

        timer_done = .false.
        nb_timer_reported = 0
        total_time_cpu = 0

        do i=1, nb_timer

            ! this a new timer that has not already been reported
            if (.not. timer_done(i)) then
                timer_done(i) = .true.

                temp_name = tab_timer(i)%name
                temp_iter = tab_timer(i)%iter
                temp_counter_self_sum_cpu = self_time_cpu(i)
                temp_counter_self_sum_elapsed = self_time_elapsed(i)

                ! if this is an upper timer, we update the total time
                if (tab_timer(i)%depth == 1) then
                    total_time_cpu = total_time_cpu + tab_timer(i)%counter_sum_cpu
                    total_time_elapsed = total_time_elapsed + tab_timer(i)%counter_sum_elapsed
                    temp_name = '**'//temp_name
                end if

                ! we search forward for timers with the same name
                do j=i+1, nb_timer
                    if (tab_timer(i)%name == tab_timer(j)%name) then
                        timer_done(j) = .true.
                        temp_iter = temp_iter + tab_timer(j)%iter
                        temp_counter_self_sum_cpu = temp_counter_self_sum_cpu + self_time_cpu(j)
                        temp_counter_self_sum_elapsed = temp_counter_self_sum_elapsed + self_time_elapsed(j)
                    end if
                end do

                ! we sort the new timer into the reported_timer_array
                j = nb_timer_reported
                do while (j>=1 .and. tab_timer_sorted(j)%counter_self_sum_cpu < temp_counter_self_sum_cpu)
                    tab_timer_sorted(j+1) = tab_timer_sorted(j)
                    j = j-1
                end do

                tab_timer_sorted(j+1)%name = temp_name
                tab_timer_sorted(j+1)%iter = temp_iter                
                tab_timer_sorted(j+1)%counter_self_sum_cpu = temp_counter_self_sum_cpu
                tab_timer_sorted(j+1)%counter_self_sum_elapsed = temp_counter_self_sum_elapsed

                nb_timer_reported = nb_timer_reported + 1
            end if
        end do

        ! we print the results which are sorted
        write(*,'(a79)')'----------------------------------------------------------------------------------'
        write(*,'(a)')'TIMING REPORT: FLAT (in sec)'
        write(*,'(a)')'Time spent in the subroutine but NOT in the nested subroutines'
        write(*,'(a79)')'----------------------------------------------------------------------------------'
        write(*,'(a16,a8,a16,a17,a9,a11)') 'name'//spaces,'#calls','cpu','elapsed','%cpu','%elapsed'
        write(*,'(a79)')'----------------------------------------------------------------------------------'


        do i=1, nb_timer_reported
            write(*,'(a16,i8,f16.2,f16.2,f10.1,f10.1)')                         &
                ' '//tab_timer_sorted(i)%name,                                  &
                tab_timer_sorted(i)%iter,                                       &
                tab_timer_sorted(i)%counter_self_sum_cpu,                       &
                tab_timer_sorted(i)%counter_self_sum_elapsed,                   &
                100*tab_timer_sorted(i)%counter_self_sum_cpu/total_time_cpu,    &
                100*tab_timer_sorted(i)%counter_self_sum_elapsed/total_time_elapsed

        end do

        write(*,'(a79)')'----------------------------------------------------------------------------------'

    end subroutine

#ifdef MPI
!***************************************************************************
    subroutine timer_report_all
!***************************************************************************
    !****f* timing/timer_report_all
    !  NAME
    !    timer_report_all
    !
    !  DESCRIPTION
    !    Print both the tree report and the flat report.
    !***

        implicit none

        call MPI_comm_rank(MPI_COMM_WORLD,rank,code)

        if (rank==0) then
            write(*,'(a79)')'----------------------------------------------------------------------------------'
            write(*,'(a)')'MPI CODE: ONLY ONE PROCESS (RANK 0) IS SHOWN BELOW'
            write(*,'(a79)')'----------------------------------------------------------------------------------'

            call timer_report_tree

            write(*,*)

            call timer_report_flat
        end if

    end subroutine


#else
!***************************************************************************
    subroutine timer_report_all
!***************************************************************************
    !****f* timing/timer_report_all
    !  NAME
    !    timer_report_all
    !
    !  DESCRIPTION
    !    Print both the tree report and the flat report.
    !***

        implicit none

        call timer_report_tree

        write(*,*)

        call timer_report_flat

    end subroutine
#endif

!***************************************************************************
    double precision function self_time_cpu(id_timer)
!***************************************************************************
    !****f* timing/self_time_cpu
    !  NAME
    !    self_time_cpu
    !
    !  DESCRIPTION
    !    Return the self-cpu-time elapsed in a timer and not in its nested timers.
    !
    !  INPUTS
    !    id_timer: id of the timer whose self cpu time is required
    !
    !  OUTPUT
    !    self_time_cpu: self cpu time spent in seconds
    !***

        implicit none

        integer, intent(in) :: id_timer
        integer :: i

        ! we determine the self-time by substracting the time of the children timers
        i = id_timer+1
        self_time_cpu = tab_timer(id_timer)%counter_sum_cpu

        do while (i<= nb_timer .and. tab_timer(i)%id_parent == id_timer)
            self_time_cpu = self_time_cpu - tab_timer(i)%counter_sum_cpu
            i = i + 1
        end do

    end function


!***************************************************************************
    double precision function self_time_elapsed(id_timer)
!***************************************************************************
    !****f* timing/self_time_elapsed
    !  NAME
    !    self_time_elapsed
    !
    !  DESCRIPTION
    !    Return the self-elapsed-time elapsed in a timer and not in its nested timers.
    !
    !  INPUTS
    !    id_timer: id of the timer whose self cpu time is required
    !
    !  OUTPUT
    !    self_time_elapsed: self elapsed time spent in seconds
    !***

        implicit none

        integer, intent(in) :: id_timer
        integer :: i

        ! we determine the self-time by substracting the time of the children timers
        i = id_timer+1
        self_time_elapsed = tab_timer(id_timer)%counter_sum_elapsed

        do while (i<= nb_timer .and. tab_timer(i)%id_parent == id_timer)
            self_time_elapsed = self_time_elapsed - tab_timer(i)%counter_sum_elapsed
            i = i + 1
        end do

    end function



!***************************************************************************
    double precision function timing_cpu()
!***************************************************************************
    !****f* timing/timing_cpu
    !  NAME
    !    timing_cpu
    !
    !  DESCRIPTION
    !    Return the cpu time in seconds.
    !
    !  OUTPUT
    !    timing_cpu: cpu-time in seconds
    !***

        implicit none

        call cpu_time(timing_cpu)

    end function

!***************************************************************************
    integer function timing_elapsed()
!***************************************************************************
    !****f* timing/timing_elapsed
    !  NAME
    !    timing_elapsed
    !
    !  DESCRIPTION
    !    Return the current period.
    !
    !  OUTPUT
    !    timing_elapsed: current period (integer)
    !***

        implicit none

        call system_clock(count=timing_elapsed)

    end function



!***************************************************************************
    subroutine timer_set_active(timer_status)
!***************************************************************************
    !****f* timing/timer_set_active
    !  NAME
    !    timer_set_active
    !
    !  DESCRIPTION
    !    Allow to switch between active (default) and inactive mode. In inactive mode, new timers are
    !    not created when calling the subroutine timer_start. It is a way of not timing some parts
    !    of the code.
    !
    !  INPUTS
    !    timer_status:  new timer status (.false. or .true.)
    !***

        implicit none

        logical, intent(in) :: timer_status

        timer_active = timer_status

    end subroutine

end module timing
